import logging
from enum import IntEnum

from ophyd import Component as Cpt
from ophyd import DerivedSignal, Device, EpicsSignal, EpicsSignalRO, PVPositioner

log = logging.getLogger(__name__)


class DoneStatus(IntEnum):
    MOVING = 0
    DONE = 1


class BusyStatus(IntEnum):
    DONE = 0
    BUSY = 1


class MotorDriveStatus(IntEnum):
    NOT_READY = 0
    READY_TO_MOVE = 1


class UndulatorPositioner(PVPositioner):
    setpoint = Cpt(EpicsSignal, "SetC.VAL")
    readback = Cpt(EpicsSignalRO, "M.VAL")

    actuate = Cpt(DerivedSignal, derived_from="parent.start_button", kind="omitted")
    stop_signal = Cpt(DerivedSignal, derived_from="parent.stop_button", kind="omitted")
    done = Cpt(DerivedSignal, derived_from="parent.done", kind="omitted")
    done_value = DoneStatus.DONE


class PlanarUndulator(Device):
    """APS Planar Undulator

    .. index:: Ophyd Device; PlanarUndulator

    The signals *busy* and *done* convey complementary
    information. *busy* comes from the IOC, while *done* comes
    directly from the controller.

    EXAMPLE::

        undulator = PlanarUndulator("S25ID:USID:", name="undulator")

    """

    # X-ray spectrum parameters
    energy = Cpt(UndulatorPositioner, "Energy")
    energy_taper = Cpt(UndulatorPositioner, "TaperEnergy")
    gap = Cpt(UndulatorPositioner, "Gap")
    gap_taper = Cpt(UndulatorPositioner, "TaperGap")
    harmonic_value = Cpt(EpicsSignal, "HarmonicValueC", kind="config")
    total_power = Cpt(EpicsSignalRO, "TotalPowerM.VAL", kind="config")
    # Signals for moving the undulator
    start_button = Cpt(EpicsSignal, "StartC.VAL", put_complete=True, kind="omitted")
    stop_button = Cpt(EpicsSignal, "StopC.VAL", kind="omitted")
    busy = Cpt(EpicsSignalRO, "BusyM.VAL", kind="omitted")
    done = Cpt(EpicsSignalRO, "BusyDeviceM.VAL", kind="omitted")
    motor_drive_status = Cpt(EpicsSignalRO, "MotorDriveStatusM.VAL", kind="omitted")
    # Miscellaneous control signals
    gap_deadband = Cpt(EpicsSignal, "DeadbandGapC", kind="config")
    device_limit = Cpt(EpicsSignal, "DeviceLimitM.VAL", kind="config")
    access_mode = Cpt(EpicsSignalRO, "AccessSecurityC", kind="omitted")
    message1 = Cpt(EpicsSignalRO, "Message1M.VAL", kind="omitted")
    message2 = Cpt(EpicsSignalRO, "Message2M.VAL", kind="omitted")
    device = Cpt(EpicsSignalRO, "DeviceM", kind="config")
    magnet = Cpt(EpicsSignalRO, "DeviceMagnetM", kind="config")
    location = Cpt(EpicsSignalRO, "LocationM", kind="config")
    version_plc = Cpt(EpicsSignalRO, "PLCVersionM.VAL", kind="config")
    version_hpmu = Cpt(EpicsSignalRO, "HPMUVersionM.VAL", kind="config")


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from unittest import mock

import pytest
from apstools.devices.aps_undulator import PlanarUndulator
from bluesky_queueserver_api import BPlan
from ophyd.sim import make_fake_device
from qtpy import QtCore

import haven
from firefly.energy import EnergyDisplay

FakeMonochromator = make_fake_device(haven.instrument.monochromator.Monochromator)
FakeEnergyPositioner = make_fake_device(
    haven.instrument.energy_positioner.EnergyPositioner
)
FakeUndulator = make_fake_device(PlanarUndulator)


@pytest.fixture()
def energy_positioner(sim_registry):
    energy = FakeEnergyPositioner(
        mono_prefix="mono_ioc:",
        undulator_prefix="id_ioc:",
        name="energy",
    )
    energy.monochromator.energy.user_setpoint.sim_set_limits((4000, 33000))
    energy.undulator.energy.setpoint.sim_set_limits((-float("inf"), float("inf")))
    return energy


@pytest.fixture()
def display(qtbot, energy_positioner):
    # Load display
    display = EnergyDisplay()
    qtbot.addWidget(display)
    return display


def test_mono_caqtdm_macros(display):
    """Example of mono caQtDM macros from microprobe mono:

    last file: /net/s25data/xorApps/ui/DCMControlCenter.ui

    macro: P=25idbUP:, MONO=UP, BRAGG=ACS:m3, GAP=ACS:m4,
    ENERGY=Energy, OFFSET=Offset, IDENERGY=ID25ds:Energy.VAL

    """
    display.launch_caqtdm = mock.MagicMock()
    # Check that the various caqtdm calls set up the right macros
    display.launch_mono_caqtdm()
    assert display.launch_caqtdm.called
    assert display.launch_caqtdm.call_args[1]["macros"] == {
        "P": "mono_ioc:",
        "MONO": "UP",
        "BRAGG": "ACS:m3",
        "GAP": "ACS:m4",
        "ENERGY": "Energy",
        "OFFSET": "Offset",
        "IDENERGY": "id_ioc:Energy",
    }


def test_id_caqtdm_macros(display):
    display.launch_caqtdm = mock.MagicMock()
    # Check that the various caqtdm calls set up the right macros
    display.launch_id_caqtdm()
    assert display.launch_caqtdm.called
    assert display.launch_caqtdm.call_args[1]["macros"] == {
        "ID": "id_ioc",
        "M": 2,
        "D": 2,
    }


def test_move_energy(qtbot, display):
    # Click the set energy button
    btn = display.ui.set_energy_button
    expected_item = BPlan("set_energy", energy=8402.0)

    def check_item(item):
        return item.to_dict() == expected_item.to_dict()

    qtbot.keyClicks(display.target_energy_lineedit, "8402")
    with qtbot.waitSignal(
        display.queue_item_submitted, timeout=1000, check_params_cb=check_item
    ):
        qtbot.mouseClick(btn, QtCore.Qt.LeftButton)


def test_predefined_energies(qtbot, display):
    # Check that the combo box was populated
    combo_box = display.ui.edge_combo_box
    assert combo_box.count() > 0
    assert combo_box.itemText(0) == "Select edge…"
    assert combo_box.itemText(1) == "Ca K (4038 eV)"
    # Does it filter energies outside the usable range?
    assert combo_box.count() < 250
    # Does it update the energy line edit?
    with qtbot.waitSignal(combo_box.activated, timeout=1000):
        qtbot.keyClicks(combo_box, "Ni K (8333 eV)\t")
        combo_box.activated.emit(9)  # <- this shouldn't be necessary
    line_edit = display.ui.target_energy_lineedit
    assert line_edit.text() == "8333.000"


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



"""The scaler has a global trigger, but individual channels.

These tests check the system for monitoring the trigger system in
bluesky so that it only fires the trigger once for all of the scaler
channels. Additionally, the scaler can send that trigger to other
devices, like the Xspress3 fluorescence detector readout electronics.

"""

import pytest
from ophyd import Device

from haven.instrument.scaler_triggered import ScalerTriggered


@pytest.mark.skip("scaler_triggering is not needed right now")
def test_trigger_fires():
    scaler_prefix = "myioc:myscaler"
    # Prepare the scaler triggered device
    MyDevice = type("MyDevice", (ScalerTriggered, Device), {})
    device = MyDevice(name="device", scaler_prefix=scaler_prefix)
    # Make sure it's in a sensible starting state
    assert hasattr(device, "_statuses")
    assert len(device._statuses) == 0
    # Trigger the device
    device.trigger()
    # Check that a status was added
    assert len(device._statuses) == 1
    # Check that triggering again is indempotent
    first_status = device._statuses[scaler_prefix]
    old_done = first_status.__class__.done
    first_status.__class__.done = False
    try:
        device.trigger()
    except Exception:
        pass
    finally:
        first_status.__class__.done = old_done
    second_status = device._statuses[scaler_prefix]
    assert first_status is second_status


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from firefly import slits
from haven.instrument import mirrors


class MirrorDisplay(slits.SlitsDisplay):
    caqtdm_ui_filenames = {
        mirrors.HighHeatLoadMirror: "/net/s25data/xorApps/epics/synApps_6_2/ioc/25ida/25idaApp/op/ui/HHLM_4.ui",
        mirrors.BendableHighHeatLoadMirror: "/net/s25data/xorApps/epics/synApps_6_2/ioc/25ida/25idaApp/op/ui/HHLM_6.ui",
    }

    def customize_ui(self):
        # Enable the bender controls if the mirror is bendable
        if self.device.bendable:
            self.ui.bender_embedded_display.setEnabled(True)

    def ui_filename(self):
        return "mirror.ui"

    def launch_caqtdm(self):
        # Sort out the prefix from the slit designator
        prefix = self.device.prefix.strip(":")
        pieces = prefix.split(":")
        # Build the macros for the caQtDM panels
        P = ":".join(pieces[:-1])
        caqtdm_macros = {
            "P": f"{P}:",
            "MIR": f"{pieces[-1]}:",
            "Y": self.device.transverse.prefix.split(":")[-1],
            "ROLL": self.device.roll.prefix.split(":")[-1],
            "LAT": self.device.normal.prefix.split(":")[-1],
            "CP": self.device.pitch.prefix.split(":")[-1],
            "UPL": self.device.upstream.prefix.split(":")[-1],
            "DNL": self.device.downstream.prefix.split(":")[-1],
        }
        if self.device.bendable:
            caqtdm_macros["BEND"] = self.device.bender.prefix.split(":")[-1]
        # Launch the caQtDM panel
        super(slits.SlitsDisplay, self).launch_caqtdm(macros=caqtdm_macros)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



"""Tests for a generic X-Y stage."""

import pytest

from haven import exceptions, registry
from haven.instrument import stage


def test_stage_init():
    stage_ = stage.XYStage(
        "motor_ioc", pv_vert=":m1", pv_horiz=":m2", labels={"stages"}, name="aerotech"
    )
    assert stage_.name == "aerotech"
    assert stage_.vert.name == "aerotech_vert"
    # Check registry of the stage and the individiual motors
    registry.clear()
    with pytest.raises(exceptions.ComponentNotFound):
        registry.findall(label="motors")
    with pytest.raises(exceptions.ComponentNotFound):
        registry.findall(label="stages")
    registry.register(stage_)
    assert len(list(registry.findall(label="motors"))) == 2
    assert len(list(registry.findall(label="stages"))) == 1


def test_load_stages(sim_registry):
    stages = stage.load_stages()
    assert len(stages) == 1
    assert isinstance(stages[0], stage.XYStage)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from haven.instrument import power_supply


def test_load_power_supplies(sim_registry):
    power_supply.load_power_supplies()
    # Test that the device has the right configuration
    devices = list(sim_registry.findall(label="power_supplies"))
    assert len(devices) == 2  # 2 channels on the device
    device = devices[0]
    assert "NHQ01_ch" in device.name


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import asyncio
import logging
import re
import time as ttime
from typing import Callable, Union

from caproto import CaprotoTimeoutError
from caproto.asyncio.client import Context
from ophyd import Component, Device, K
from ophyd.sim import make_fake_device

from .._iconfig import load_config

log = logging.getLogger(__name__)


async def resolve_device_names(ic_defns):
    """Update ion chamber definitions to include the EPICS name field.

    Updates *ic_defns* in place to add the *name* key. Each entry in
    *ic_defns* should have a *desc_pv* key with the PV to be checked
    for the device name.

    Ensures names are safe as Ophyd device names.

    """

    async def get_name(pv):
        try:
            response = await pv.read()
        except CaprotoTimeoutError:
            return None
        else:
            # Read PV values over the network
            return response.data.tobytes().strip(b"\x00").decode("latin-1")

    ctx = Context()
    desc_pvs = await ctx.get_pvs(*[defn["desc_pv"] for defn in ic_defns])
    names = await asyncio.gather(*(get_name(pv) for pv in desc_pvs))
    # Add the results back into the defitions
    for name, defn in zip(names, ic_defns):
        defn["name"] = name


def titelize(name):
    """Convert a device name into a human-readable title."""
    title = name.replace("_", " ").title()
    # Replace select phrases that are known to be incorrect
    replacements = {"Kb ": "KB "}
    for orig, new in replacements.items():
        title = title.replace(orig, new)
    return title


async def aload_devices(*coros):
    return await asyncio.gather(*coros)


def make_device(DeviceClass, *args, FakeDeviceClass=None, **kwargs) -> Device:
    """Create device and add it to the registry.

    If the beamline is not connected, i.e. the config file has:

        [beamline]
        is_connected = false

    then the created device will be simulated.

    Parameters
    ==========
    DeviceClass
      The device class to use for making this device.
    FakeDeviceClass
      If the beamline is not connected, use this device instead.

    Returns
    =======
    device
      The newly created and registered camera object.

    """
    # Make a fake device if the beamline is not connected
    config = load_config()
    if config["beamline"]["is_connected"]:
        Cls = DeviceClass
    else:
        # Make fake device
        if FakeDeviceClass is None:
            Cls = make_fake_device(DeviceClass)
        else:
            Cls = FakeDeviceClass
    # Make sure we can connect
    name = kwargs.get("name", "unknown")
    t0 = ttime.monotonic()
    device = Cls(
        *args,
        **kwargs,
    )
    return device


async def await_for_connection(dev, all_signals=False, timeout=3.0):
    """Wait for signals to connect

    Parameters
    ----------
    all_signals : bool, optional
        Wait for all signals to connect (including lazy ones)
    timeout : float or None
        Overall timeout
    """
    signals = [walk.item for walk in dev.walk_signals(include_lazy=all_signals)]

    pending_funcs = {
        dev: getattr(dev, "_required_for_connection", {})
        for name, dev in dev.walk_subdevices(include_lazy=all_signals)
    }
    pending_funcs[dev] = dev._required_for_connection

    t0 = ttime.monotonic()
    # Wait until all the signals have connected
    loop_idx = 0
    connected = False
    while True:
        # Check if the device is ready
        if not connected:
            connected = all(sig.connected for sig in signals)
        if connected and not any(pending_funcs.values()):
            return
        # Since we're not connected, sleep for a short time and try again
        real_timeout = min((0.05, timeout / 10.0))
        tn = ttime.monotonic()
        await asyncio.sleep(real_timeout)
        # Detect other co-routines blocking for too long
        tdelay = ttime.monotonic() - tn
        if tdelay > (real_timeout * 5):
            msg = f"{real_timeout} sec sleep for {dev.name} took "
            msg += f"{tdelay:.4f} sec. "
            msg += "Maybe another co-routine is blocking."
            log.info(msg)
        elapsed_time = ttime.monotonic() - t0
        loop_idx += 1
        if timeout is not None and elapsed_time > timeout and loop_idx > 2:
            break

    def get_name(sig):
        sig_name = f"{dev.name}.{sig.dotted_name}"
        return f"{sig_name} ({sig.pvname})" if hasattr(sig, "pvname") else sig_name

    reasons = []
    unconnected = ", ".join(get_name(sig) for sig in signals if not sig.connected)
    if unconnected:
        reasons.append(f"Failed to connect to all signals: {unconnected}")
    if any(pending_funcs.values()):
        pending = ", ".join(
            description.format(device=dev)
            for dev, funcs in pending_funcs.items()
            for obj, description in funcs.items()
        )
        reasons.append(f"Pending operations: {pending}")
    raise TimeoutError(dev.name + "; ".join(reasons))


class RegexComponent(Component[K]):
    r"""A component with regular expression matching.

    In EPICS, it is not possible to add a field to an existing record,
    e.g. adding a ``.RnXY`` field to go alongside ``mca1.RnNM`` and
    other fields in the MCA record. A common solution is to create a
    new record with an underscore instead of the dot: ``mca1_RnBH``.

    This component include these types of field-like-records as part
    of the ROI device with a ``mca1.Rn`` prefix but performing
    subsitution on the device name using regular expressions. See the
    documentation for ``re.sub`` for full details.

    Example
    =======

    .. code:: python

        class ROI(mca.ROI):
            name = RECpt(EpicsSignal, "NM", lazy=True)
            is_hinted = RECpt(EpicsSignal, "BH",
                              pattern=r"^(.+)\.R(\d+)",
                              repl=r"\1_R\2",
                              lazy=True)

        class MCA(mca.EpicsMCARecord):
            roi0 = Cpt(ROI, ".R0")
            roi1 = Cpt(ROI, ".R1")

        mca = MCA(prefix="mca")
        # *name* has the normal concatination
        assert mca.roi0.name.pvname == "mca.R0NM"
        # *is_hinted* has regex substitution
        assert mca.roi0.is_hinted.pvname == "mca_R0BH"

    """

    def __init__(self, *args, pattern: str, repl: Union[str, Callable], **kwargs):
        """*pattern* and *repl* match their use in ``re.sub``."""
        self.pattern = pattern
        self.repl = repl
        super().__init__(*args, **kwargs)

    def maybe_add_prefix(self, instance, kw, suffix):
        """Parse prefix and suffix with regex suffix if kw is in self.add_prefix.

        Parameters
        ----------
        instance : Device
            The instance from which to extract the prefix to maybe
            append to the suffix.

        kw : str
            The key of associated with the suffix.  If this key is
            self.add_prefix than prepend the prefix to the suffix and
            return, else just return the suffix.

        suffix : str
            The suffix to maybe have something prepended to.

        Returns
        -------
        str

        """
        new_val = super().maybe_add_prefix(instance, kw, suffix)
        try:
            new_val = re.sub(self.pattern, self.repl, new_val)
        except TypeError:
            pass
        return new_val


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import time
from collections import OrderedDict

import numpy as np
import pytest
from ophyd import ADComponent as ADCpt
from ophyd.areadetector.cam import AreaDetectorCam
from ophyd.sim import instantiate_fake_device

from haven.instrument.area_detector import (
    DetectorBase,
    DetectorState,
    HDF5FilePlugin,
    load_area_detectors,
)


class Detector(DetectorBase):
    cam = ADCpt(AreaDetectorCam, "cam1:")
    hdf = ADCpt(HDF5FilePlugin, "HDF1:", write_path_template="/tmp/")


@pytest.fixture()
def detector(sim_registry):
    det = instantiate_fake_device(Detector)
    return det


def test_flyscan_kickoff(detector):
    detector.flyer_num_points.set(10)
    status = detector.kickoff()
    detector.cam.detector_state.sim_put(DetectorState.ACQUIRE)
    status.wait(timeout=3)
    assert status.success
    assert status.done
    # Check that the device was properly configured for fly-scanning
    assert detector.cam.acquire.get() == 1
    assert detector._fly_data == {}
    # Check that timestamps get recorded when new data are available
    detector.cam.array_counter.sim_put(1)
    event = detector._fly_data[detector.cam.array_counter]
    assert event[0].timestamp == pytest.approx(time.time())


def test_flyscan_complete(sim_ion_chamber):
    flyer = sim_ion_chamber
    # Run the complete method
    status = flyer.complete()
    status.wait(timeout=3)
    # Check that the detector is stopped
    assert flyer.stop_all._readback == 1


def test_flyscan_collect(sim_ion_chamber):
    flyer = sim_ion_chamber
    name = flyer.net_counts.name
    flyer.start_timestamp = 988.0
    # Make fake fly-scan data
    sim_data = np.zeros(shape=(8000,))
    sim_data[:6] = [3, 5, 8, 13, 2, 33]
    flyer.mca.spectrum._readback = sim_data
    sim_times = np.asarray([12.0e7, 4.0e7, 4.0e7, 4.0e7, 4.0e7, 4.0e7])
    flyer.mca_times.spectrum._readback = sim_times
    flyer.frequency.set(1e7).wait(timeout=3)
    # Ignore the first collected data point because it's during taxiing
    expected_data = sim_data[1:]
    # The real timestamps should be midway between PSO pulses
    flyer.timestamps = [1000, 1004, 1008, 1012, 1016, 1020]
    expected_timestamps = [1002.0, 1006.0, 1010.0, 1014.0, 1018.0]
    payload = list(flyer.collect())
    # Confirm data have the right structure
    for datum, value, timestamp in zip(payload, expected_data, expected_timestamps):
        assert datum == {
            "data": {name: [value]},
            "timestamps": {name: [timestamp]},
            "time": timestamp,
        }


def test_load_area_detectors(sim_registry):
    load_area_detectors()
    # Check that some area detectors were loaded
    dets = sim_registry.findall(label="area_detectors")


def test_hdf_dtype(detector):
    """Check that the right ``dtype_str`` is added to the image data to
    make tiled happy.
    """
    # Set up fake image metadata
    detector.hdf.data_type.sim_put("UInt8")
    original_desc = OrderedDict(
        [
            (
                "FakeDetector_image",
                {
                    "shape": (1, 1024, 1280),
                    "source": "PV:25idcARV4:",
                    "dtype": "array",
                    "external": "FILESTORE:",
                },
            )
        ]
    )
    # Update and check the description
    new_desc = detector.hdf._add_dtype_str(original_desc)
    assert new_desc["FakeDetector_image"]["dtype_str"] == "|u1"


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from apstools.synApps import TransformRecord
from ophyd import Component as Cpt
from ophyd import Device, Kind, OphydObject

from .. import exceptions
from .._iconfig import load_config
from .device import make_device
from .motor import HavenMotor


class Table(Device):
    """A dynamic Ophyd device for an optical table.

    This device will have the components needed for the configuration
    of motors provided. For example, if *horizontal_motor* is
    provided, then the device will have a ``horizontal`` attribute
    that is the corresponding horizontal motor.

    There are several dynamic parameters when creating an object that
    control the structure of the resulting device. These are described
    individually below and are assumed relative to *prefix*. If
    *upstream_motor* and *downstream_motor* are not empty, then
    *vertical* and *pitch* will point to a pseudo motor who's PV is
    determined by *pseudo_motors* (the value of *vertical_motor* will
    be ignored in this case). For example:

    .. code:: python

        tbl = Table(
            "255idcVME:",
            pseudo_motors="table_us:",
            transforms="table_us_trans:",
            upstream_motor="m3",
            downstream_motor="m4",
            name="my_table",
        )

    will produce a table with the following components:

    - *upstream*: "255idcVME:m3"
    - *downstream*: "255idcVME:m4"
    - *vertical*: "255idcVME:table_us:height"
    - *pitch*: "255idcVME:table_us:pitch"
    - *drive*: transform "255idcVME:table_us_trans:Drive"
    - *readback*: transform "255idcVME:table_us_trans:Readback"

    Parameters
    ==========
    prefix
      The PV prefix common to all components. E.g. "25idcVME:".
    vertical_motor
      The suffix for the real motor controlling the height of the
      table (e.g. "m1"). Ignored if *upstream_motor* and
      *downstream_motor* are provided.
    horizontal_motor
      The suffix for the real motor controlling the lateral movement
      of the table (e.g. "m1").
    upstream_motor
      The suffix for the real motor controlling the height of the
      upstream leg of the table (e.g. "m1").
    downstream_motor
      The suffix for the real motor controlling the height of the
      downstream leg of the table (e.g. "m1").
    pseudo_motors
      The suffix for the pseudo motors controlling the orientation of
      the table. E.g. "table_ds:". Creates the components: *height*
      and *pitch*.
    transforms
      The suffix for the pseudo motor transform records
      (e.g. "table_ds_trans:"). Creates the components:
      *vertical_drive_transform" and "vertical_readback_transform".

    """

    # These are the possible components that could be present
    vertical: OphydObject
    horizontal: OphydObject
    upstream: OphydObject
    downstream: OphydObject
    pitch: OphydObject
    horizontal_drive_transform: OphydObject
    horizontal_readback_transform: OphydObject

    def __new__(
        cls,
        prefix,
        *args,
        vertical_motor: str = "",
        horizontal_motor: str = "",
        upstream_motor: str = "",
        downstream_motor: str = "",
        pseudo_motors: str = "",
        transforms: str = "",
        **kwargs,
    ):
        # See if we have direct control over the vertical/horizontal positions
        comps = {}
        if bool(vertical_motor):
            comps["vertical"] = Cpt(
                HavenMotor,
                vertical_motor,
                labels={"motors"},
            )
        if bool(horizontal_motor):
            comps["horizontal"] = Cpt(
                HavenMotor,
                horizontal_motor,
                labels={"motors"},
            )
        if bool(upstream_motor):
            comps["upstream"] = Cpt(
                HavenMotor,
                upstream_motor,
                labels={"motors"},
            )
        if bool(downstream_motor):
            comps["downstream"] = Cpt(
                HavenMotor,
                downstream_motor,
                labels={"motors"},
            )
        # Check if we need to add the pseudo motors and tranforms
        if bool(upstream_motor) and bool(downstream_motor):
            comps["vertical"] = Cpt(
                HavenMotor, f"{pseudo_motors}height", labels={"motors"}
            )
            comps["pitch"] = Cpt(HavenMotor, f"{pseudo_motors}pitch", labels={"motors"})
            comps["vertical_drive_transform"] = Cpt(
                TransformRecord, f"{transforms}Drive", kind=Kind.config
            )
            comps["vertical_readback_transform"] = Cpt(
                TransformRecord, f"{transforms}Readback", kind=Kind.config
            )
        # Now create a customized class for all the motors given
        new_cls = type("Table", (cls,), comps)
        return object.__new__(new_cls)

    def __init__(
        self,
        *args,
        pseudo_motors: str = "",
        transforms: str = "",
        vertical_motor: str = "",
        horizontal_motor: str = "",
        upstream_motor: str = "",
        downstream_motor: str = "",
        **kwargs,
    ):
        # __init__ needs to accept the arguments accepted by __new__.
        self.pseudo_motors = pseudo_motors
        self.transforms = transforms
        super().__init__(*args, **kwargs)


def load_tables(config=None):
    if config is None:
        config = load_config()
    # Create two-bounce KB mirror sets
    devices = []
    for name, tbl_config in config.get("table", {}).items():
        # Build the motor prefixes
        try:
            prefix = tbl_config["prefix"]
            attr_names = [
                "upstream_motor",
                "downstream_motor",
                "horizontal_motor",
                "vertical_motor",
                "transforms",
                "pseudo_motors",
            ]
            attrs = {attr: tbl_config.get(attr, "") for attr in attr_names}
        except KeyError as ex:
            raise exceptions.UnknownDeviceConfiguration(
                f"Device {name} missing '{ex.args[0]}': {tbl_config}"
            ) from ex
        # Make the device
        devices.append(
            make_device(
                Table,
                prefix=prefix,
                name=name,
                labels={"tables"},
                **attrs,
            )
        )
    return devices


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import datetime as dt
import logging
import os
import time
from collections import ChainMap
from io import StringIO
from pathlib import Path

import numpy as np
import pytest

# from freezegun import freeze_time
import pytz
import time_machine
from bluesky import RunEngine
from numpy import asarray as array
from ophyd.sim import SynAxis, SynGauss, motor

from haven import XDIWriter, energy_scan, exceptions

fake_time = pytz.timezone("America/New_York").localize(
    dt.datetime(2022, 8, 19, 19, 10, 51)
)


log = logging.getLogger(__name__)
# logging.basicConfig(level=logging.DEBUG)


# Stub the epics signals in Haven

THIS_DIR = Path(__file__).parent


# Sample metadata dict
{
    "E0": 0,
    "beamline": {
        "is_connected": False,
        "name": "SPC Beamline (sector unknown)",
        "pv_prefix": "",
    },
    "detectors": ["I0", "It"],
    "edge": "Ni_K",
    "facility": {"name": "Advanced Photon Source", "xray_source": "insertion device"},
    "hints": {"dimensions": [(["energy", "exposure"], "primary")]},
    "ion_chambers": {"scaler": {"pv_prefix": ""}},
    "motors": ["energy", "exposure"],
    "num_intervals": 9,
    "num_points": 10,
    "plan_args": {
        "args": [
            (
                "SynAxis(prefix='', name='energy', "
                "read_attrs=['readback', 'setpoint'], "
                "configuration_attrs=['velocity', 'acceleration'])"
            ),
            array([8300, 8310, 8320, 8330, 8340, 8350, 8360, 8370, 8380, 8390]),
            (
                "SynAxis(prefix='', name='exposure', "
                "read_attrs=['readback', 'setpoint'], "
                "configuration_attrs=['velocity', 'acceleration'])"
            ),
            [0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1],
        ],
        "detectors": [
            (
                "SynGauss(prefix='', name='I0', "
                "read_attrs=['val'], configuration_attrs=['Imax', "
                "'center', 'sigma', 'noise', 'noise_multiplier'])"
            ),
            (
                "SynGauss(prefix='', name='It', "
                "read_attrs=['val'], configuration_attrs=['Imax', "
                "'center', 'sigma', 'noise', "
                "'noise_multiplier'])"
            ),
        ],
        "per_step": "None",
    },
    "plan_name": "list_scan",
    "plan_pattern": "inner_list_product",
    "plan_pattern_args": {
        "args": [
            (
                "SynAxis(prefix='', name='energy', "
                "read_attrs=['readback', 'setpoint'], "
                "configuration_attrs=['velocity', "
                "'acceleration'])"
            ),
            array([8300, 8310, 8320, 8330, 8340, 8350, 8360, 8370, 8380, 8390]),
            (
                "SynAxis(prefix='', name='exposure', "
                "read_attrs=['readback', 'setpoint'], "
                "configuration_attrs=['velocity', "
                "'acceleration'])"
            ),
            [0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1],
        ]
    },
    "plan_pattern_module": "bluesky.plan_patterns",
    "plan_type": "generator",
    "sample_name": "NiO_rock_salt",
    "scan_id": 1,
    "time": 1661025010.409619,
    "uid": "671c3c48-f014-421d-b3e0-57991b6745f6",
    "versions": {"bluesky": "1.8.3", "ophyd": "1.6.4"},
}


start_doc = {
    "versions": {"bluesky": "1.8.3", "ophyd": "1.6.4"},
    "detectors": ["I0", "It"],
    "d_spacing": 3.0,
    "motors": ["energy", "exposure"],
    "edge": "Ni_K",
    "facility": {
        "name": "Advanced Photon Source",
        "xray_source": "insertion device",
    },
    "beamline": {"name": "20-ID-C", "pv_prefix": "20id:"},
    "sample_name": "nickel oxide",
    "uid": "671c3c48-f014-421d-b3e0-57991b6745f6",
}
event_doc = {
    "data": {
        "I0_net_counts": 2,
        "It_net_counts": 1.5,
        "IpreKB_net_counts": 2.5,
        "Ipreslit_net_counts": 2.2,
        "Iref_net_counts": 0.56,
        "energy": 8330,
        "exposure": 0.1,
        "sim motor 1": 8330.0,
    },
    "time": 1660186828.0055554,
    "descriptor": "7ed5b8c5-045c-41fa-b79c-50fbcbe777e5",
}
descriptor_doc = {
    "hints": {
        "I0": {"fields": ["I0_net_counts"]},
        "IpreKB": {"fields": ["IpreKB_net_counts"]},
        "Ipreslit": {"fields": ["Ipreslit_net_counts"]},
        "Iref": {"fields": ["Iref_net_counts"]},
        "It": {"fields": ["It_net_counts"]},
        "sim motor 1": {"fields": ["sim motor 1"]},
        "energy": {"fields": ["energy"]},
    },
    "name": "primary",
    "run_start": "6974290f-fe3f-4535-bb8c-c29c915f88aa",
    "time": 1712865522.9266968,
    "uid": "7ed5b8c5-045c-41fa-b79c-50fbcbe777e5",
}


@pytest.fixture()
def stringio():
    yield StringIO()


@pytest.fixture()
def file_path(tmp_path):
    fp = tmp_path / "sample_file.txt"
    try:
        yield fp
    finally:
        fp.unlink()


@pytest.fixture()
def writer(stringio):
    yield XDIWriter(stringio)


def test_opens_file(file_path):
    # Pass in a string and it should hold the path until the start document is found
    fp = file_path
    writer = XDIWriter(fp)
    assert writer.fp == fp
    assert not fp.exists()
    # Run the writer
    writer("start", {"edge": "Ni_K"})
    # Check that a file was created
    assert fp.exists()


def test_uses_open_file(stringio):
    # Pass in a string and it should hold the path until the start document is found
    writer = XDIWriter(stringio)
    assert writer.fd is stringio


def test_read_only_file(file_path):
    # Check that it raises an exception if an open file has no write intent
    # Put some content in the temporary file
    with open(file_path, mode="w") as fd:
        fd.write("Hello, spam!")
    # Check that the writer raises an exception when the file is open read-only
    with open(file_path, mode="r") as fd:
        with pytest.raises(exceptions.FileNotWritable):
            XDIWriter(fd)


def test_required_headers(writer):
    writer("start", start_doc)
    writer("descriptor", descriptor_doc)
    writer("event", event_doc)  # Header gets written on first event
    # Check that required headers were added to the XDI file
    writer.fd.seek(0)
    xdi_output = writer.fd.read()
    assert "# XDI/1.0 bluesky/1.8.3 ophyd/1.6.4" in xdi_output
    assert "# Column.1: energy" in xdi_output
    assert "# Element.symbol: Ni" in xdi_output
    assert "# Element.edge: K" in xdi_output
    assert "# Mono.d_spacing: 3" in xdi_output  # Not implemented yet
    assert "# -------------" in xdi_output


@time_machine.travel(fake_time)
def test_optional_headers(writer):
    os.environ["TZ"] = "America/New_York"
    time.tzset()
    writer("start", start_doc)
    writer("descriptor", descriptor_doc)
    writer("event", event_doc)
    # Check that required headers were added to the XDI file
    writer.fd.seek(0)
    xdi_output = writer.fd.read()
    expected_metadata = {
        "Facility.name": "Advanced Photon Source",
        "Facility.xray_source": "insertion device",
        "Beamline.name": "20-ID-C",
        "Beamline.pv_prefix": "20id:",
        "Scan.start_time": "2022-08-19 19:10:51-0400",
        "Column.8": "time",
        "uid": "671c3c48-f014-421d-b3e0-57991b6745f6",
    }
    for key, val in expected_metadata.items():
        assert f"# {key.lower()}: {val.lower()}\n" in xdi_output.lower()


@time_machine.travel(fake_time)
def test_file_path_formatting(tmp_path):
    """Check that "{date}_{user}.xdi" formatting works in the filename."""
    writer = XDIWriter(tmp_path / "{year}{month}{day}_{short_uid}_{sample_name}.xdi")
    writer.start(start_doc)
    assert str(writer.fp) == str(tmp_path / "20220819_671c3c48_nickel-oxide.xdi")


@time_machine.travel(fake_time)
def test_file_path_reentry(tmp_path):
    """Check that "{date}_{user}.xdi" formatting can be used multiple times."""
    # Start the writer once with basic arguments
    writer = XDIWriter(tmp_path / "{year}{month}{day}_{short_uid}_{sample_name}.xdi")
    writer.start(start_doc)
    target_path = str(tmp_path / "20220819_671c3c48_nickel-oxide.xdi")
    assert str(writer.fp) == target_path
    assert writer.fd.name == target_path
    # Check that additional start docs don't create new files
    writer.start(start_doc)
    # Start the writer again with a second set of arguments
    new_start_doc = ChainMap(
        {
            "sample_name": "manganese oxide",
            "uid": "a6842eaa-6dd3-4666-83a0-1829cd687556",
        },
        start_doc,
    )
    writer.start(new_start_doc)
    target_path = str(tmp_path / "20220819_a6842eaa_manganese-oxide.xdi")
    assert str(writer.fp) == target_path
    assert writer.fd.name == target_path


@time_machine.travel(fake_time)
def test_secondary_stream(tmp_path):
    """Check that secondary data streams get ignored."""
    sec_event = ChainMap(
        {
            "descriptor": "b1006389-fd92-4037-9eb3-02332703552b",
            "data": {"Iref": 2},
        },
        event_doc,
    )
    sec_descriptor = ChainMap(
        {"uid": sec_event["descriptor"], "name": "secondary"}, descriptor_doc
    )
    # Set up the writer
    writer = XDIWriter(tmp_path / "{year}{month}{day}_{short_uid}_{sample_name}.xdi")
    writer.start(start_doc)
    # Events before the descriptor should raise exceptions
    assert writer._primary_uid == None
    with pytest.raises(exceptions.DocumentNotFound):
        writer.event(event_doc)
    # Prime the writer with descriptor documents
    writer.descriptor(descriptor_doc)
    writer.descriptor(sec_descriptor)
    # Send a correct primary data event
    writer.event(event_doc)
    # Send an event from secondary data stream
    writer.event(sec_event)


@time_machine.travel(fake_time)
def test_manager_path(tmp_path, beamline_manager):
    """Check that "{manager_path}.xdi" formatting works in the filename."""
    # Set up a full path on the beamline manager
    beamline_manager.local_storage.full_path._readback = str(tmp_path) + "/"
    # Create the XDI writer object
    writer = XDIWriter(
        "{manager_path}/{year}{month}{day}_{short_uid}_{sample_name}.xdi"
    )
    writer.start(start_doc)
    assert str(writer.fp) == f"{tmp_path}/20220819_671c3c48_nickel-oxide.xdi"


def test_file_path_formatting_bad_key():
    """Check that "{date}_{user}.xdi" raises exception if placeholder is invalid."""
    writer = XDIWriter("{year}{month}{day}_{spam}_{sample_name}.xdi")
    with pytest.raises(exceptions.XDIFilenameKeyNotFound):
        writer.start(start_doc)


def test_data(writer):
    """Check that the TSV data section is present and correct."""
    writer.start(start_doc)
    writer.descriptor(descriptor_doc)
    writer("event", event_doc)
    # Verify the data were written properly
    writer.fd.seek(0)
    xdi_output = writer.fd.read()
    assert "8330" in xdi_output
    assert "1660186828.0055554" in xdi_output


def test_with_plan(stringio, sim_registry, event_loop, beamline_manager, tmp_path):
    beamline_manager.local_storage.full_path._readback = str(tmp_path) + "/"
    I0 = SynGauss(
        "I0",
        motor,
        "motor",
        center=-0.5,
        Imax=1,
        sigma=1,
        labels={"detectors"},
    )
    It = SynGauss(
        "It",
        motor,
        "motor",
        center=-0.5,
        Imax=1,
        sigma=1,
        labels={"detectors"},
    )
    energy_motor = SynAxis(name="energy", labels={"motors", "energies"})
    exposure = SynAxis(name="exposure", labels={"motors", "exposures"})
    writer = XDIWriter(stringio)
    RE = RunEngine()
    energy_plan = energy_scan(
        np.arange(8300.0, 8400.0, 10),
        detectors=[I0, It],
        E0="Ni_K",
        energy_signals=[energy_motor],
        time_signals=[exposure],
        md=dict(sample_name="NiO_rock_salt"),
    )
    RE(energy_plan, writer)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import logging
import warnings
from typing import Mapping, Sequence

from apstools.utils.misc import safe_ophyd_name
from ophyd import Component as Cpt
from ophyd import EpicsMotor, EpicsSignal, EpicsSignalRO, Kind

from .._iconfig import load_config
from .device import make_device, resolve_device_names
from .instrument_registry import InstrumentRegistry
from .instrument_registry import registry as default_registry
from .motor_flyer import MotorFlyer

log = logging.getLogger(__name__)


class HavenMotor(MotorFlyer, EpicsMotor):
    """The default motor for haven movement.

    This motor also implements the flyer interface and so can be used
    in a fly scan, though no hardware triggering is supported.

    Returns to the previous value when being unstaged.

    """

    # Extra motor record components
    encoder_resolution = Cpt(EpicsSignal, ".ERES", kind=Kind.config)
    description = Cpt(EpicsSignal, ".DESC", kind="omitted")
    tweak_value = Cpt(EpicsSignal, ".TWV", kind="omitted")
    tweak_forward = Cpt(EpicsSignal, ".TWF", kind="omitted", tolerance=2)
    tweak_reverse = Cpt(EpicsSignal, ".TWR", kind="omitted", tolerance=2)
    motor_stop = Cpt(EpicsSignal, ".STOP", kind="omitted", tolerance=2)
    soft_limit_violation = Cpt(EpicsSignalRO, ".LVIO", kind="omitted")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def stage(self):
        super().stage()
        # Override some additional staged signals
        self._original_vals.setdefault(self.user_setpoint, self.user_readback.get())
        self._original_vals.setdefault(self.velocity, self.velocity.get())


async def load_motors(
    config: Mapping = None, registry: InstrumentRegistry = default_registry
) -> Sequence:
    """Load generic hardware motors from IOCs.

    This loader will skip motor prefixes that already exist in the
    registry *registry*, so it is a good idea to run this loader after
    other devices have been created that might potentially use some of
    these motors (e.g. mirrors, tables, etc.).

    Parameters
    ==========
    config
      The beamline configuration. If omitted, will use the config
      provided by :py:func:`haven._iconfig.load_config()`.
    registry
      The instrument registry to check for existing motors. Existing
      motors will not be duplicated.

    Returns
    =======
    devices
      The newly create EpicsMotor devices.

    """
    if config is None:
        config = load_config()
    # Build up definitions of motors to load
    defns = []
    for section_name, config in config.get("motor", {}).items():
        prefix = config["prefix"]
        num_motors = config["num_motors"]
        log.info(
            f"Preparing {num_motors} motors from IOC: " f"{section_name} ({prefix})"
        )
        for idx in range(num_motors):
            motor_prefix = f"{prefix}m{idx+1}"
            defns.append(
                {
                    "prefix": motor_prefix,
                    "desc_pv": f"{motor_prefix}.DESC",
                    "ioc_name": section_name,
                }
            )
    # Check that we're not duplicating a motor somewhere else (e.g. KB mirrors)
    existing_pvs = []
    for m in registry.findall(label="motors", allow_none=True):
        if hasattr(m, "prefix"):
            existing_pvs.append(m.prefix)
    defns = [defn for defn in defns if defn["prefix"] not in existing_pvs]
    duplicates = [defn for defn in defns if defn["prefix"] in existing_pvs]
    if len(duplicates) > 0:
        log.info(
            "The following motors already exist and will not be duplicated: ",
            ", ".join([m["prefix"] for m in duplicates]),
        )
    else:
        log.debug(f"No duplicated motors detected out of {len(defns)}")
    # Resolve the scaler channels into ion chamber names
    await resolve_device_names(defns)
    # Create the devices
    devices = []
    missing_channels = []
    unnamed_channels = []
    for defn in defns:
        # Check for motor without a name
        if defn["name"] == "":
            unnamed_channels.append(defn["prefix"])
        elif defn["name"] is None:
            missing_channels.append(defn["prefix"])
        else:
            # Create the device
            labels = {"motors", "extra_motors", "baseline", defn["ioc_name"]}
            name = safe_ophyd_name(defn["name"])
            devices.append(
                make_device(HavenMotor, prefix=defn["prefix"], name=name, labels=labels)
            )
    # Notify about motors that have no name
    if len(missing_channels) > 0:
        msg = "Skipping unavailable motors: "
        msg += ", ".join([prefix for prefix in missing_channels])
        warnings.warn(msg)
        log.warning(msg)
    if len(unnamed_channels) > 0:
        msg = "Skipping unnamed motors: "
        msg += ", ".join([prefix for prefix in unnamed_channels])
        warnings.warn(msg)
        log.warning(msg)
    return devices


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import qtawesome as qta
from qtpy import QtWidgets

from firefly import display
from haven import registry


class IonChamberDisplay(display.FireflyDisplay):
    """A GUI window for changing settings in an ion chamber."""

    caqtdm_scaler_ui_file: str = "/net/s25data/xorApps/ui/scaler32_full_offset.ui"
    caqtdm_mcs_ui_file: str = (
        "/APSshare/epics/synApps_6_2_1/support/mca-R7-9//mcaApp/op/ui/autoconvert/SIS38XX.ui"
    )
    caqtdm_preamp_ui_file: str = (
        "/net/s25data/xorApps/epics/synApps_6_2_1/support/ip-GIT/ipApp/op/ui/autoconvert/SR570.ui"
    )

    def customize_device(self):
        self._device = registry.find(self.macros()["IC"])

    def customize_ui(self):
        # Use qtawesome icons instead of unicode arrows
        self.ui.gain_down_button.setText("")
        self.ui.gain_down_button.setIcon(qta.icon("fa5s.arrow-left"))
        self.ui.gain_up_button.setText("")
        self.ui.gain_up_button.setIcon(qta.icon("fa5s.arrow-right"))

    def ui_filename(self):
        return "ion_chamber.ui"

    def prepare_caqtdm_actions(self):
        self.caqtdm_actions = []
        # Create an action for launching the scaler caQtDM file
        action = QtWidgets.QAction(self)
        action.setObjectName("launch_scaler_caqtdm_action")
        action.setText("&Scaler caQtDM")
        action.triggered.connect(self.launch_scaler_caqtdm)
        action.setIcon(qta.icon("fa5s.wrench"))
        action.setToolTip("Launch the caQtDM panel for the scaler.")
        self.caqtdm_actions.append(action)
        # Create an action for launching the MCS caQtDM file
        action = QtWidgets.QAction(self)
        action.setObjectName("launch_mcs_caqtdm_action")
        action.setText("&MCS caQtDM")
        action.triggered.connect(self.launch_mcs_caqtdm)
        action.setIcon(qta.icon("fa5s.wrench"))
        action.setToolTip(
            "Launch the caQtDM panel for the multi-channel scaler controls."
        )
        self.caqtdm_actions.append(action)
        # Create an action for launching the Preamp caQtDM file
        action = QtWidgets.QAction(self)
        action.setObjectName("launch_preamp_caqtdm_action")
        action.setText("&Preamp caQtDM")
        action.triggered.connect(self.launch_preamp_caqtdm)
        action.setIcon(qta.icon("fa5s.wrench"))
        action.setToolTip("Launch the caQtDM panel for the preamplifier.")
        self.caqtdm_actions.append(action)

    def launch_scaler_caqtdm(self):
        device = self._device
        caqtdm_macros = {
            "P": f"{device.scaler_prefix}:",
            "S": "scaler1",
        }
        return self.launch_caqtdm(
            macros=caqtdm_macros, ui_file=self.caqtdm_scaler_ui_file
        )

    def launch_mcs_caqtdm(self):
        device = self._device
        caqtdm_macros = {
            "P": f"{device.scaler_prefix}:",
        }
        return self.launch_caqtdm(macros=caqtdm_macros, ui_file=self.caqtdm_mcs_ui_file)

    def launch_preamp_caqtdm(self):
        device = self._device
        sep = ":"
        bits = device.preamp_prefix.strip(sep).split(sep)
        prefix = sep.join(bits[:-1])
        amp_suffix = bits[-1]
        caqtdm_macros = {
            "P": f"{prefix}:",
            "A": f"{amp_suffix}:",
        }
        return self.launch_caqtdm(
            macros=caqtdm_macros, ui_file=self.caqtdm_preamp_ui_file
        )


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import logging
from typing import Mapping

from ophyd import Component as Cpt
from ophyd import FormattedComponent as FCpt
from ophyd import PVPositionerPC
from pcdsdevices.signal import MultiDerivedSignal

from .._iconfig import load_config
from .device import make_device
from .monochromator import Monochromator
from .xray_source import PlanarUndulator

log = logging.getLogger(__name__)


__all__ = ["EnergyPositioner", "load_energy_positioner"]


class EnergyPositioner(PVPositionerPC):
    """The operational energy of the beamline.

    Responsible for setting both mono and ID energy with an optional
    ID offset. Setting the *energy* component will propagate to both
    real devices and so ``EnergyPositioner().energy`` is a good
    candidate for a positioner for Bluesky plans.

    Currently, the offset between the insertion device and the
    monochromator is fixed. In the future this will be replaced with a
    more sophisticated calculation.

    .. todo::

       Insert functionality to have a non-constant ID offset.

    Attributes
    ==========

    id_offset
      The offset for the insertion device relative to the mono energy.
    energy
      The pseudo positioner for the forward calculation.
    mono_energy
      The real component for the monochromator energy.
    id_energy
      The real component for the insertion device energy.

    Parameters
    ==========
    mono_prefix
      The CA prefix for the monochromator IOC.
    undulator_prefix
      The prefix for the insertion device energy, such that
      f"{id_prefix}:Energy.VAL" reaches the energy readback value.

    """

    # Individual energy components
    monochromator = FCpt(Monochromator, "{mono_prefix}", labels={"monochromators"})
    undulator = FCpt(
        PlanarUndulator, "{undulator_prefix}", labels={"undulator", "xray_source"}
    )

    def __init__(
        self,
        mono_prefix: str,
        undulator_prefix: str,
        *args,
        **kwargs,
    ):
        self.mono_prefix = mono_prefix
        self.undulator_prefix = undulator_prefix
        super().__init__(*args, **kwargs)

    def set_energy(self, *, mds: MultiDerivedSignal, value: float):
        ev_per_kev = 1000
        offset = self.monochromator.id_offset.get()
        vals = {
            self.monochromator.energy: value,
            self.undulator.energy: (value + offset) / ev_per_kev,
        }
        return vals

    def get_energy(self, mds: MultiDerivedSignal, items: Mapping):
        if self.monochromator.energy.user_readback in items:
            energy = items[self.monochromator.energy.user_readback]
        else:
            energy = items[self.monochromator.energy.user_setpoint]
        return energy

    @property
    def limits(self):
        hi, low = (None, None)
        for signal in self.setpoint.signals:
            # Update the limits based on this signal
            try:
                new_low, new_hi = signal.limits
            except TypeError:
                continue
            # Account for the keV -> eV conversion for the undulator
            if signal is self.undulator.energy.setpoint:
                new_low *= 1000
                new_hi *= 1000
            hi = min([val for val in (hi, new_hi) if val is not None])
            low = max([val for val in (low, new_low) if val is not None])
        return (low, hi)

    setpoint = Cpt(
        MultiDerivedSignal,
        attrs={"monochromator.energy.user_setpoint", "undulator.energy.setpoint"},
        calculate_on_get=get_energy,
        calculate_on_put=set_energy,
        name="setpoint",
    )
    readback = Cpt(
        MultiDerivedSignal,
        attrs={"monochromator.energy.user_readback"},
        calculate_on_get=get_energy,
        calculate_on_put=set_energy,
        name="readback",
    )


def load_energy_positioner(config=None):
    # Load PV's from config
    if config is None:
        config = load_config()
    # Guard to make sure we have a mono and ID configuration
    if "monochromator" not in config.keys() or "undulator" not in config.keys():
        return
    # Make the combined energy device
    device = make_device(
        EnergyPositioner,
        name="energy",
        mono_prefix=config["monochromator"]["prefix"],
        undulator_prefix=config["undulator"]["prefix"],
    )
    return device


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from ophyd import EpicsSignal, EpicsSignalRO

__all__ = ["ScalerTriggered"]


class ScalerTriggered:
    """A mix-in for triggering a device using the scaler.

    The device does not have to be a channel on the scaler, enabling
    hardware triggering of other devices.

    If this is a used as part of a component, then triggering is left
    up to the parent device.

    If this is used as part of a top-level Device, then the device
    should have a *scaler_prefix* attribute with the PV prefix to the
    scaler (e.g. "25idcVME:scaler1"), though this is only used to
    coordinate statuses for multiple devices using the same scaler.

    """

    scaler_prefix = None
    _statuses = {}

    def trigger(self, *args, **kwargs):
        is_top_device = getattr(self, "parent", None) is None
        if is_top_device:
            # This is the top-level device, so trigger it
            return self._trigger_scaler(*args, **kwargs)
        else:
            # This is a sub-component of a device, so trigger the parent
            return self.parent.trigger()

    def _trigger_scaler(self, *args, **kwargs):
        # Figure out if there's already a trigger active
        previous_status = self._statuses.get(self.scaler_prefix)
        is_idle = previous_status is None or previous_status.done
        # Trigger the detector if not already running, and update the status dict
        if is_idle:
            new_status = super().trigger(*args, **kwargs)
            self._statuses[self.scaler_prefix] = new_status
        else:
            new_status = previous_status
        return new_status


class ScalerSignal(ScalerTriggered, EpicsSignal): ...


class ScalerSignalRO(ScalerTriggered, EpicsSignalRO): ...


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



"""A PyDM data plugin that uses an Ophyd registry object to
communicate with signals.

Provides funcitonality so that PyDM channels can be addressed as e.g.
``haven://mirror.pitch.user_readback``

"""

from typhos.plugins.core import SignalConnection, SignalPlugin

from haven import registry


class HavenConnection(SignalConnection):
    def find_signal(self, address: str):
        """Find a signal in the registry given its address.
        This method is intended to be overridden by subclasses that
        may use a different mechanism to keep track of signals.
        Parameters
        ----------
        address
          The connection address for the signal. E.g. in
          "sig://sim_motor.user_readback" this would be the
          "sim_motor.user_readback" portion.
        Returns
        -------
        Signal
          The Ophyd signal corresponding to the address.
        """
        return registry[address]


class HavenPlugin(SignalPlugin):
    protocol = "haven"
    connection_class = HavenConnection



#!/usr/bin/env python3
from caproto.server import PVGroup, ioc_arg_parser, pvproperty, run

from haven.simulated_ioc import ResponsiveMotorFields


class MotorGroup(PVGroup):
    """
    An IOC with some motor records, similar to those found in a VME crate.

    E.g. "25idcVME:m1.VAL"

    """

    m1 = pvproperty(value=5000.0, doc="SLT V Upper", record=ResponsiveMotorFields)
    m2 = pvproperty(value=5000.0, doc="SLT V Lower", record=ResponsiveMotorFields)
    m3 = pvproperty(value=5000.0, doc="SLT H Inb", record=ResponsiveMotorFields)
    m4 = pvproperty(value=5000.0, doc="SLT H Otb", record=ResponsiveMotorFields)


if __name__ == "__main__":
    ioc_options, run_options = ioc_arg_parser(
        default_prefix="225idVME:", desc="haven.tests.ioc_motor test IOC"
    )
    ioc = MotorGroup(**ioc_options)
    run(ioc.pvdb, **run_options)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from firefly import display


class TweakDisplay(display.FireflyDisplay):
    """A small set of widgets for tweaking a signal by specified value."""

    def customize_ui(self):
        if hasattr(self.ui, "value_spin_box"):
            self.ui.value_spin_box.valueChanged.connect(self.update_press_values)

    def update_press_values(self, tweak_value):
        """Update the press values for the buttons.

        This makes them move in increments of the tweak value.

        """
        self.ui.reverse_button.pressValue = -tweak_value
        self.ui.forward_button.pressValue = tweak_value

    def ui_filename(self):
        return "tweak.ui"



import pytest
from ophyd import DetectorBase
from ophyd.sim import instantiate_fake_device

from haven import load_config, registry
from haven.instrument.camera import AravisDetector, load_cameras

PREFIX = "255idgigeA:"


def test_load_cameras(sim_registry):
    load_cameras(config=load_config())
    # Check that cameras were registered
    cameras = list(registry.findall(label="cameras"))
    assert len(cameras) == 1
    assert isinstance(cameras[0], DetectorBase)


@pytest.fixture()
def camera(sim_registry):
    camera = instantiate_fake_device(
        AravisDetector, prefix="255idgigeA:", name="camera"
    )
    return camera


def test_camera_device(camera):
    assert isinstance(camera, DetectorBase)
    assert hasattr(camera, "cam")


def test_camera_in_registry(sim_registry, camera):
    # Check that all sub-components are accessible
    sim_registry.find(f"{camera.name}_cam")
    sim_registry.find(f"{camera.name}_cam.gain")


def test_default_time_signal(camera):
    assert camera.default_time_signal is camera.cam.acquire_time


def test_hdf5_write_path(camera):
    # The HDF5 file should get dynamically re-written based on config file
    assert camera.hdf.write_path_template == "/tmp/%Y/%m/camera"


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from unittest import mock

import pytest
from ophyd.sim import make_fake_device

from firefly.table import TableDisplay
from haven.instrument import Table

FakeTable = make_fake_device(Table)


@pytest.fixture()
def table(sim_registry):
    """A fake set of slits using the 4-blade setup."""
    FakeTable = make_fake_device(Table)
    tbl = FakeTable(
        prefix="255idc:",
        name="table",
        upstream_motor="m1",
        downstream_motor="m2",
        horizontal_motor="m3",
        vertical_motor="m2",
        pseudo_motors="table_ds:",
        transforms="table_ds_trans:",
        labels={"tables"},
    )
    return tbl


@pytest.fixture()
def empty_table(sim_registry):
    """A fake set of slits using the 4-blade setup."""
    tbl = FakeTable(
        prefix="255idc:",
        name="table",
        labels={"tables"},
    )
    return tbl


@pytest.fixture()
def display(qtbot, table):
    disp = TableDisplay(macros={"DEVICE": table.name})
    qtbot.addWidget(disp)
    return disp


def test_unused_motor_widgets(qtbot, empty_table):
    """Do control widgets get disable for motors that aren't on the device?"""
    display = TableDisplay(macros={"DEVICE": empty_table.name})
    qtbot.addWidget(display)
    # Check that the bender control widgets were enabled
    assert not display.ui.pitch_embedded_display.isEnabled()
    assert not display.ui.vertical_embedded_display.isEnabled()
    assert not display.ui.horizontal_embedded_display.isEnabled()


def test_tilting_table_caqtdm(qtbot, sim_registry):
    table = FakeTable(
        "255idcVME:",
        horizontal_motor="m4",
        upstream_motor="m42",
        downstream_motor="m43",
        transforms="table_ds_trans:",
        pseudo_motors="table_ds:",
        name="table",
    )
    display = TableDisplay(macros={"DEVICE": table.name})
    qtbot.addWidget(display)
    display._open_caqtdm_subprocess = mock.MagicMock()
    # Launch the caqtdm display
    display.launch_caqtdm()
    assert display._open_caqtdm_subprocess.called
    cmds = display._open_caqtdm_subprocess.call_args[0][0]
    # Check that the right macros are sent
    # /net/s25data/xorApps/ui/table_2leg.ui
    # macro: P=25idcVME:,PM=25idcVME:,TB=table_ds,TR=table_ds_trans,TBUS=m42,TBDS=m43,TBH=m4
    # Check that the right UI file is being used
    ui_file = cmds[-1]
    assert ui_file.split("/")[-1] == "table_2leg.ui"
    # Check the macros
    macros = [cmds[i + 1] for i in range(len(cmds)) if cmds[i] == "-macro"][0]
    assert "P=255idcVME:" in macros
    assert "PM=255idcVME:" in macros
    assert "TB=table_ds" in macros
    assert "TR=table_ds_trans" in macros
    assert "TBUS=m42" in macros
    assert "TBDS=m43" in macros
    assert "TBH=m4" in macros


def test_single_motor_caqtdm(qtbot, sim_registry):
    table = FakeTable("255idcVME:", horizontal_motor="m3", name="table")
    display = TableDisplay(macros={"DEVICE": table.name})
    qtbot.addWidget(display)
    display._open_caqtdm_subprocess = mock.MagicMock()
    # Launch the caqtdm display
    display.launch_caqtdm()
    assert display._open_caqtdm_subprocess.called
    cmds = display._open_caqtdm_subprocess.call_args[0][0]
    # Check that the right macros are sent
    # /APSshare/epics/synApps_6_2_1/support/motor-R7-2-2//motorApp/op/ui/autoconvert/motorx.ui
    # macro: P=25iddVME:,M=m45
    # Check that the right UI file is being used
    ui_file = cmds[-1]
    assert ui_file.split("/")[-1] == "motorx.ui"
    # Check the macros
    macros = [cmds[i + 1] for i in range(len(cmds)) if cmds[i] == "-macro"][0]
    assert "P=255idcVME:" in macros
    assert "M=m3" in macros


def test_double_motor_caqtdm(qtbot, sim_registry):
    table = FakeTable(
        "255idcVME:", horizontal_motor="m3", vertical_motor="m4", name="table"
    )
    display = TableDisplay(macros={"DEVICE": table.name})
    qtbot.addWidget(display)
    display._open_caqtdm_subprocess = mock.MagicMock()
    # Launch the caqtdm display
    display.launch_caqtdm()
    assert display._open_caqtdm_subprocess.called
    cmds = display._open_caqtdm_subprocess.call_args[0][0]
    # Check that the right macros are sent
    # /APSshare/epics/synApps_6_2_1/support/motor-R7-2-2//motorApp/op/ui/autoconvert/motor2x.ui,
    # macro: P=25idcVME:,M1=m9,M2=m10
    # Check that the right UI file is being used
    ui_file = cmds[-1]
    assert ui_file.split("/")[-1] == "motor2x.ui"
    # Check the macros
    macros = [cmds[i + 1] for i in range(len(cmds)) if cmds[i] == "-macro"][0]
    assert "P=255idcVME:" in macros
    assert "M1=m3" in macros
    assert "M2=m4" in macros


# def test_kb_bendable_mirrors_caqtdm(ffapp, kb_bendable_mirrors):
#     mirrors = kb_bendable_mirrors
#     disp = KBMirrorsDisplay(macros={"DEVICE": mirrors.name})
#     disp._open_caqtdm_subprocess = mock.MagicMock()
#     # Launch the caqtdm display
#     disp.launch_caqtdm()
#     assert disp._open_caqtdm_subprocess.called
#     cmds = disp._open_caqtdm_subprocess.call_args[0][0]
#     # Check that the right macros are sent
#     # /net/s25data/xorApps/ui/KB_mirrors_and_benders.ui, macro: P=25idcVME:,PM=25idcVME:,KB=LongKB_Cdn,KBH=LongKB_Cdn:H,KBV=LongKB_Cdn:V,KBHUS=m50,KBHDS=m49,KBVUS=m56,KBVDS=m55,KB1=LongKB_CdnH, KB2=LongKB_CdnV,HBUS=m52,HBDS=m51,VBUS=m54,VBDS=m53
#     macros = [cmds[i + 1] for i in range(len(cmds)) if cmds[i] == "-macro"][0]
#     assert "P=255idc:" in macros
#     assert "PM=255idc:" in macros
#     assert "KB=Long_KB" in macros
#     assert "KBH=Long_KB:H" in macros
#     assert "KBV=Long_KB:V" in macros
#     assert "KBHUS=m5" in macros
#     assert "KBHDS=m7" in macros
#     assert "KBVUS=m6" in macros
#     assert "KBVDS=m8" in macros
#     assert "KB1=Long_KBH" in macros
#     assert "KB2=Long_KBV" in macros
#     assert "HBUS=m21" in macros
#     assert "HBDS=m23" in macros
#     assert "VBUS=m22" in macros
#     assert "VBDS=m24" in macros
#     # Check that the right UI file is being used
#     ui_file = cmds[-1]
#     assert ui_file.split("/")[-1] == "KB_mirrors_and_benders.ui"



import logging
from enum import IntEnum

from ophyd import Component as Cpt
from ophyd import Device, EpicsSignal, EpicsSignalRO

from .motor import HavenMotor

log = logging.getLogger(__name__)


class IDTracking(IntEnum):
    OFF = 0
    ON = 1


class Monochromator(Device):
    # ID tracking PVs
    id_tracking = Cpt(EpicsSignal, "ID_tracking", kind="config")
    id_offset = Cpt(EpicsSignal, "ID_offset", kind="config")
    d_spacing = Cpt(EpicsSignal, "dspacing", kind="config")
    # Virtual positioners
    mode = Cpt(EpicsSignal, "mode", labels={"motors", "baseline"}, kind="config")
    energy = Cpt(HavenMotor, "Energy", labels={"motors"}, kind="hinted")
    energy_constant1 = Cpt(
        EpicsSignal, "EnergyC1.VAL", labels={"baseline"}, kind="config"
    )
    energy_constant2 = Cpt(
        EpicsSignal, "EnergyC2.VAL", labels={"baseline"}, kind="config"
    )
    energy_constant3 = Cpt(
        EpicsSignal, "EnergyC3.VAL", labels={"baseline"}, kind="config"
    )
    offset = Cpt(HavenMotor, "Offset", labels={"motors", "baseline"}, kind="config")
    # ACS Motors
    horiz = Cpt(HavenMotor, "ACS:m1", labels={"motors", "baseline"}, kind="config")
    vert = Cpt(HavenMotor, "ACS:m2", labels={"motors", "baseline"}, kind="config")
    bragg = Cpt(HavenMotor, "ACS:m3", labels={"motors"})
    gap = Cpt(HavenMotor, "ACS:m4", labels={"motors"})
    roll2 = Cpt(HavenMotor, "ACS:m5", labels={"motors", "baseline"}, kind="config")
    pitch2 = Cpt(HavenMotor, "ACS:m6", labels={"motors", "baseline"}, kind="config")
    # roll_int = Cpt(HavenMotor, "ACS:m7", labels={"motors", "baseline"}, kind="config")
    # pi_int = Cpt(HavenMotor, "ACS:m8", labels={"motors", "baseline"}, kind="config")
    # Physical constants
    d_spacing = Cpt(EpicsSignalRO, "dspacing", labels={"baseline"}, kind="config")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.stage_sigs[self.id_tracking] = IDTracking.OFF


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import pydm
from ophyd import Signal
from pydm.data_plugins import plugin_for_address
from pydm.widgets import PyDMLineEdit

from firefly.pydm_plugin import HavenPlugin


def test_plugin_registered():
    plugin = plugin_for_address("haven://")
    assert isinstance(plugin, HavenPlugin)


def test_signal_connection(qapp, qtbot, sim_registry):
    # Create a signal and attach our listener
    sig = Signal(name="my_signal", value=1)
    widget = PyDMLineEdit()
    qtbot.addWidget(widget)
    widget.channel = "haven://my_signal"
    listener = widget.channels()[0]
    # If PyDMChannel can not connect, we need to connect it ourselves
    # In PyDM > 1.5.0 this will not be neccesary as the widget will be
    # connected after we set the channel name
    if not hasattr(listener, "connect"):
        pydm.utilities.establish_widget_connections(widget)
    # Check that our widget receives the initial value
    qapp.processEvents()
    # breakpoint()
    assert widget._write_access
    assert widget._connected
    assert widget.value == 1



from unittest.mock import MagicMock

import pytest
from ophyd import sim

from haven import mono_ID_calibration


@pytest.fixture()
def mono_motor(sim_registry):
    motor = sim.SynAxis(name="energy_mono_energy", labels={"motors", "energies"})
    yield motor


@pytest.fixture()
def pitch2_motor(sim_registry):
    motor = sim.SynAxis(name="monochromator_pitch2", labels={"motors"})
    yield motor


@pytest.fixture()
def id_motor(sim_registry):
    motor = sim.SynAxis(name="energy_id_energy", labels={"motors", "energies"})
    yield motor


@pytest.fixture()
def ion_chamber(sim_registry, id_motor):
    I0 = sim.SynGauss(
        "I0",
        id_motor,
        "energy_id_energy",
        center=8.0,
        Imax=1,
        sigma=1,
        labels={"ion_chambers"},
    )
    yield I0


@pytest.mark.skip(reason="``haven.plans.align_motor`` is deprecated.")
def test_moves_energy(mono_motor, id_motor, ion_chamber, pitch2_motor, event_loop, RE):
    """Simple test to ensure that the plan moves the mono and undulators
    to the right starting energy."""
    # Execute the plan
    id_motor.set(8)
    fit_model = MagicMock()
    RE(
        mono_ID_calibration(
            energies=[8000], energy_motor=mono_motor, fit_model=fit_model
        )
    )
    assert mono_motor.readback.get() == 8000.0


def test_aligns_pitch(mono_motor, id_motor, ion_chamber, pitch2_motor):
    fit_model = MagicMock()
    plan = mono_ID_calibration(
        energies=[8000], energy_motor=mono_motor, fit_model=fit_model
    )
    with pytest.warns(UserWarning):
        # Raises a warning because there's no data to fit
        messages = list(plan)
    device_names = [getattr(m.obj, "name", None) for m in messages]
    assert "monochromator_pitch2" in device_names


def test_aligns_mono_energy(mono_motor, id_motor, ion_chamber, pitch2_motor):
    fit_model = MagicMock()
    plan = mono_ID_calibration(
        energies=[8000], energy_motor=mono_motor, fit_model=fit_model
    )
    with pytest.warns(UserWarning):
        # Raises a warning because there's no data to fit
        messages = list(plan)
    id_messages = [
        m for m in messages if getattr(m.obj, "name", "") == "energy_id_energy"
    ]
    # Check that messages were produced for each point in the mono scan
    npts = 40  # Taken from `align_motor` plan
    assert len(id_messages) >= npts


@pytest.mark.skip(reason="``haven.plans.align_motor`` is deprecated.")
def test_fitting_callback(
    mono_motor, id_motor, ion_chamber, pitch2_motor, event_loop, RE
):
    fit_model = MagicMock()
    plan = mono_ID_calibration(
        energies=[8000, 9000], energy_motor=mono_motor, fit_model=fit_model
    )
    # Execute the plan in the runengine
    result = RE(plan)
    # Check that the fitting results are available
    fit_model.fit.assert_called_once()
    # assert False


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from apstools.synApps import TransformRecord
from ophyd import Component as Cpt
from ophyd import Device
from ophyd import FormattedComponent as FCpt
from ophyd import Kind

from .. import exceptions
from .._iconfig import load_config
from .device import RegexComponent as RCpt
from .device import make_device
from .motor import HavenMotor


class HighHeatLoadMirror(Device):
    bendable = False
    # Physical motors
    transverse = Cpt(HavenMotor, "m1", labels={"motors"})
    roll = Cpt(HavenMotor, "m2", labels={"motors"})
    upstream = Cpt(HavenMotor, "m3", labels={"motors"})
    downstream = Cpt(HavenMotor, "m4", labels={"motors"})

    # Pseudo motors
    pitch = Cpt(HavenMotor, "coarsePitch", kind=Kind.hinted, labels={"motors"})
    normal = Cpt(HavenMotor, "lateral", kind=Kind.hinted, labels={"motors"})

    # Standard transform records for the pseudo motors
    drive_transform = Cpt(TransformRecord, "lats:Drive", kind=Kind.config)
    readback_transform = Cpt(TransformRecord, "lats:Readback", kind=Kind.config)


class BendableHighHeatLoadMirror(HighHeatLoadMirror):
    bendable = True
    bender = Cpt(HavenMotor, "m5", labels={"motors"})


class KBMirror(Device):
    """A single mirror in a KB mirror set."""

    bendable = False

    pitch = Cpt(HavenMotor, "pitch", labels={"motors"})
    normal = Cpt(HavenMotor, "height", labels={"motors"})
    upstream = FCpt(HavenMotor, "{upstream_motor}", labels={"motors"})
    downstream = FCpt(HavenMotor, "{downstream_motor}", labels={"motors"})

    # The pseudo motor transform records have
    # a missing ':', so we need to remove it.
    drive_transform = RCpt(
        TransformRecord, "Drive", pattern=":([HV]):", repl=r"\1:", kind=Kind.config
    )
    readback_transform = RCpt(
        TransformRecord, "Readback", pattern=":([HV]):", repl=r"\1:", kind=Kind.config
    )

    def __init__(
        self,
        *args,
        upstream_motor: str,
        downstream_motor: str,
        upstream_bender: str = "",
        downstream_bender: str = "",
        **kwargs,
    ):
        self.upstream_motor = upstream_motor
        self.downstream_motor = downstream_motor
        self._upstream_bender = upstream_bender
        self._downstream_bender = downstream_bender
        super().__init__(*args, **kwargs)


class BendableKBMirror(KBMirror):
    """A single bendable mirror in a KB mirror set."""

    bendable = True

    bender_upstream = FCpt(HavenMotor, "{_upstream_bender}", labels={"motors"})
    bender_downstream = FCpt(HavenMotor, "{_downstream_bender}", labels={"motors"})


class KBMirrors(Device):
    def __new__(
        cls,
        *args,
        horiz_upstream_motor: str,
        horiz_downstream_motor: str,
        vert_upstream_motor: str,
        vert_downstream_motor: str,
        horiz_upstream_bender: str = "",
        horiz_downstream_bender: str = "",
        vert_upstream_bender: str = "",
        vert_downstream_bender: str = "",
        **kwargs,
    ):
        # Decide if the mirrors are bendable or not
        HorizClass = VertClass = KBMirror
        if bool(horiz_upstream_bender) and bool(horiz_downstream_bender):
            HorizClass = BendableKBMirror
        if bool(vert_upstream_bender) and bool(vert_downstream_bender):
            VertClass = BendableKBMirror
        # Create a customized subclass based on the configuration attrs
        attrs = dict(
            horiz=Cpt(
                HorizClass,
                "H:",
                upstream_motor=horiz_upstream_motor,
                downstream_motor=horiz_downstream_motor,
                upstream_bender=horiz_upstream_bender,
                downstream_bender=horiz_downstream_bender,
            ),
            vert=Cpt(
                VertClass,
                "V:",
                upstream_motor=vert_upstream_motor,
                downstream_motor=vert_downstream_motor,
                upstream_bender=vert_upstream_bender,
                downstream_bender=vert_downstream_bender,
            ),
        )
        NewMirrors = type("KBMirrors", (cls,), attrs)
        return super().__new__(NewMirrors)

    def __init__(
        self,
        *args,
        horiz_upstream_motor: str,
        horiz_downstream_motor: str,
        vert_upstream_motor: str,
        vert_downstream_motor: str,
        horiz_upstream_bender: str = "",
        horiz_downstream_bender: str = "",
        vert_upstream_bender: str = "",
        vert_downstream_bender: str = "",
        **kwargs,
    ):
        super().__init__(*args, **kwargs)


def load_mirrors(config=None):
    if config is None:
        config = load_config()
    # Create two-bounce KB mirror sets
    devices = []
    for name, kb_config in config.get("kb_mirrors", {}).items():
        # Build the motor prefixes
        try:
            prefix = kb_config["prefix"]
            ioc_prefix = prefix.split(":")[0]
            motors = dict(
                # Normal motors
                horiz_upstream_motor=kb_config["horiz_upstream_motor"],
                horiz_downstream_motor=kb_config["horiz_downstream_motor"],
                vert_upstream_motor=kb_config["vert_upstream_motor"],
                vert_downstream_motor=kb_config["vert_downstream_motor"],
                # Bender motors
                horiz_upstream_bender=kb_config.get("horiz_upstream_bender", ""),
                horiz_downstream_bender=kb_config.get("horiz_downstream_bender", ""),
                vert_upstream_bender=kb_config.get("vert_upstream_bender", ""),
                vert_downstream_bender=kb_config.get("vert_downstream_bender", ""),
            )
            # Convert motors to fully qualified PV names (if not empty)
            motors = {
                key: f"{ioc_prefix}:{val}" for key, val in motors.items() if bool(val)
            }
        except KeyError as ex:
            raise exceptions.UnknownDeviceConfiguration(
                f"Device {name} missing '{ex.args[0]}': {kb_config}"
            )
        # Make the device
        devices.append(
            make_device(
                KBMirrors, prefix=prefix, name=name, labels={"kb_mirrors"}, **motors
            )
        )
    # Create single-bounce mirrors
    for name, mirror_config in config.get("mirrors", {}).items():
        # Decide which base class of mirror to use
        class_name = mirror_config["device_class"]
        if mirror_config.get("bendable", False):
            class_name = f"Bendable{class_name}"
        DeviceClass = globals().get(class_name)
        # Check that it's a valid device class
        if DeviceClass is None:
            msg = f"mirrors.{name}.device_class={mirror_config['device_class']}"
            raise exceptions.UnknownDeviceConfiguration(msg)
        devices.append(
            make_device(
                DeviceClass,
                prefix=mirror_config["prefix"],
                name=name,
                labels={"mirrors"},
            )
        )
    return devices


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import enum

from ophyd import Component as Cpt
from ophyd import Device, EpicsSignal, EpicsSignalRO, Kind


class EpicsSignalWithIO(EpicsSignal):
    # An EPICS signal that simply uses the DG-645 convention of
    # 'AO' being the setpoint and 'AI' being the read-back

    def __init__(self, prefix, **kwargs):
        super().__init__(f"{prefix}I", write_pv=f"{prefix}O", **kwargs)


class DG645Channel(Device):
    reference = Cpt(EpicsSignalWithIO, "ReferenceM", kind=Kind.config)
    delay = Cpt(EpicsSignalWithIO, "DelayA", kind=Kind.config)


class DG645Output(Device):
    output_mode_ttl = Cpt(EpicsSignal, "OutputModeTtlSS.PROC", kind=Kind.config)
    output_mode_nim = Cpt(EpicsSignal, "OutputModeNimSS.PROC", kind=Kind.config)
    polarity = Cpt(EpicsSignalWithIO, "OutputPolarityB", kind=Kind.config)
    amplitude = Cpt(EpicsSignalWithIO, "OutputAmpA", kind=Kind.config)
    offset = Cpt(EpicsSignalWithIO, "OutputOffsetA", kind=Kind.config)


class DG645DelayOutput(DG645Output):
    trigger_prescale = Cpt(EpicsSignalWithIO, "TriggerPrescaleL", kind=Kind.config)
    trigger_phase = Cpt(EpicsSignalWithIO, "TriggerPhaseL", kind=Kind.config)


class DG645Delay(Device):
    label = Cpt(EpicsSignal, "Label", kind=Kind.config)
    status = Cpt(EpicsSignalRO, "StatusSI", kind=Kind.omitted)
    clear_error = Cpt(EpicsSignal, "StatusClearBO", kind=Kind.omitted)
    device_id = Cpt(EpicsSignalRO, "IdentSI", kind=Kind.omitted)
    goto_remote = Cpt(EpicsSignal, "GotoRemoteBO", kind=Kind.omitted)
    goto_local = Cpt(EpicsSignal, "GotoLocalBO", kind=Kind.omitted)
    reset = Cpt(EpicsSignal, "ResetBO", kind=Kind.omitted)
    status_checking = Cpt(EpicsSignal, "StatusCheckingBO", kind=Kind.omitted)
    reset_serial = Cpt(EpicsSignal, "IfaceSerialResetBO", kind=Kind.omitted)
    serial_state = Cpt(EpicsSignalWithIO, "IfaceSerialStateB", kind=Kind.omitted)
    serial_baud = Cpt(EpicsSignalWithIO, "IfaceSerialBaudM", kind=Kind.omitted)
    reset_gpib = Cpt(EpicsSignal, "IfaceGpibResetBO", kind=Kind.omitted)
    gpib_state = Cpt(EpicsSignalWithIO, "IfaceGpibStateB", kind=Kind.omitted)
    gpib_address = Cpt(EpicsSignalWithIO, "IfaceGpibAddrL", kind=Kind.omitted)
    reset_lan = Cpt(EpicsSignal, "IfaceLanResetBO", kind=Kind.omitted)
    mac_address = Cpt(EpicsSignalRO, "IfaceMacAddrSI", kind=Kind.omitted)
    lan_state = Cpt(EpicsSignalWithIO, "IfaceLanStateB", kind=Kind.omitted)
    dhcp_state = Cpt(EpicsSignalWithIO, "IfaceDhcpStateB", kind=Kind.omitted)
    autoip_state = Cpt(EpicsSignalWithIO, "IfaceAutoIpStateB", kind=Kind.omitted)
    static_ip_state = Cpt(EpicsSignalWithIO, "IfaceStaticIpStateB", kind=Kind.omitted)
    bare_socket_state = Cpt(
        EpicsSignalWithIO, "IfaceBareSocketStateB", kind=Kind.omitted
    )
    telnet_state = Cpt(EpicsSignalWithIO, "IfaceTelnetStateB", kind=Kind.omitted)
    vxi11_state = Cpt(EpicsSignalWithIO, "IfaceVxiStateB", kind=Kind.omitted)
    ip_address = Cpt(EpicsSignalWithIO, "IfaceIpAddrS", kind=Kind.omitted)
    network_mask = Cpt(EpicsSignalWithIO, "IfaceNetMaskS", kind=Kind.omitted)
    gateway = Cpt(EpicsSignalWithIO, "IfaceGatewayS", kind=Kind.omitted)

    # Individual delay channels
    channel_A = Cpt(DG645Channel, "A")
    channel_B = Cpt(DG645Channel, "B")
    channel_C = Cpt(DG645Channel, "C")
    channel_D = Cpt(DG645Channel, "D")
    channel_E = Cpt(DG645Channel, "E")
    channel_F = Cpt(DG645Channel, "F")
    channel_G = Cpt(DG645Channel, "G")
    channel_H = Cpt(DG645Channel, "H")

    # 2-channel delay outputs
    output_T0 = Cpt(DG645Output, "T0")
    output_AB = Cpt(DG645DelayOutput, "AB")
    output_CD = Cpt(DG645DelayOutput, "CD")
    output_EF = Cpt(DG645DelayOutput, "EF")
    output_GH = Cpt(DG645DelayOutput, "GH")

    # Trigger control
    trigger_source = Cpt(EpicsSignalWithIO, "TriggerSourceM", kind=Kind.config)
    trigger_inhibit = Cpt(EpicsSignalWithIO, "TriggerInhibitM", kind=Kind.config)
    trigger_level = Cpt(EpicsSignalWithIO, "TriggerLevelA", kind=Kind.config)
    trigger_rate = Cpt(EpicsSignalWithIO, "TriggerRateA", kind=Kind.config)
    trigger_advanced_mode = Cpt(
        EpicsSignalWithIO, "TriggerAdvancedModeB", kind=Kind.config
    )
    trigger_holdoff = Cpt(EpicsSignalWithIO, "TriggerHoldoffA", kind=Kind.config)
    trigger_prescale = Cpt(EpicsSignalWithIO, "TriggerPrescaleL", kind=Kind.config)

    # Burst settings
    burst_mode = Cpt(EpicsSignalWithIO, "BurstModeB", kind=Kind.config)
    burst_count = Cpt(EpicsSignalWithIO, "BurstCountL", kind=Kind.config)
    burst_mode = Cpt(EpicsSignalWithIO, "BurstConfigB", kind=Kind.config)
    burst_delay = Cpt(EpicsSignalWithIO, "BurstDelayA", kind=Kind.config)
    burst_period = Cpt(EpicsSignalWithIO, "BurstPeriodA", kind=Kind.config)

    class trigger_sources(enum.IntEnum):
        INTERNAL = 0
        EXT_RISING_EDGE = 1
        EXT_FALLING_EDGE = 2
        SS_EXT_RISE_EDGE = 3
        SS_EXT_FALL_EDGE = 4
        SINGLE_SHOT = 5
        LINE = 6

    class polarities(enum.IntEnum):
        NEGATIVE = 0
        POSITIVE = 1

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.stage_sigs[self.trigger_source] = self.trigger_sources.EXT_RISING_EDGE


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import time

import numpy as np
import pytest

from haven.instrument import ion_chamber


@pytest.fixture()
def preamp(sim_ion_chamber):
    preamp = sim_ion_chamber.preamp
    return preamp


def test_get_gain_level(preamp):
    # Change the preamp settings
    preamp.sensitivity_value.put("20")
    assert preamp.sensitivity_value.get(as_string=True) == "20"
    assert preamp.sensitivity_value.get(as_string=False) == 4
    preamp.sensitivity_unit.put("uA/V"),
    preamp.offset_value.put(1),  # 2 uA/V
    preamp.offset_unit.put(2),
    # Check that the gain level moved
    assert preamp.gain_level.get(use_monitor=False) == 5


def test_put_gain_level(preamp):
    # Move the gain level
    preamp.gain_level.set(15).wait(timeout=3)
    # Check that the preamp sensitivities are moved
    assert preamp.sensitivity_value.get(use_monitor=False) == "10"
    assert preamp.sensitivity_unit.get(use_monitor=False) == "nA/V"
    # Check that the preamp sensitivity offsets are moved
    assert preamp.offset_value.get(use_monitor=False) == "1"
    assert preamp.offset_unit.get(use_monitor=False) == "nA"


def test_gain_signals(preamp):
    # Change the preamp settings
    preamp.sensitivity_value.put("20")
    preamp.sensitivity_unit.put("uA/V")
    preamp.offset_value.put("2")
    preamp.offset_unit.put("uA")
    # Check the gain and gain_db signals
    assert preamp.gain.get(use_monitor=False) == pytest.approx(1 / 20e-6)
    assert preamp.gain_db.get(use_monitor=False) == pytest.approx(46.9897)


@pytest.mark.asyncio
async def test_load_ion_chambers(sim_registry, mocker):
    async def resolve_device_names(defns):
        for defn in defns:
            defn["name"] = f"ion_chamber_{defn['ch_num']}"

    mocker.patch("haven.ion_chamber.resolve_device_names", new=resolve_device_names)
    await ion_chamber.load_ion_chambers()
    # Test the channel info is extracted properly
    ic = sim_registry.find(label="ion_chambers")
    assert ic.ch_num == 2
    assert ic.preamp.prefix.strip(":").split(":")[-1] == "SR03"
    assert ic.voltmeter.prefix == "255idc:LabjackT7_0:Ai1"
    assert ic.counts_per_volt_second == 1e7


def test_default_pv_prefix():
    """Check that it uses the *prefix* argument if no *scaler_prefix* is
    given.

    """
    prefix = "myioc:myscaler"
    # Instantiate the device with *scaler_prefix* argument
    device = ion_chamber.IonChamber(
        name="device1", prefix="gibberish", ch_num=1, scaler_prefix=prefix
    )
    device.scaler_prefix = prefix
    assert device.scaler_prefix == prefix
    # Instantiate the device with *scaler_prefix* argument
    device = ion_chamber.IonChamber(name="device2", ch_num=1, prefix=prefix)
    assert device.scaler_prefix == prefix


def test_volts_signal(sim_ion_chamber):
    """Test that the scaler tick counts get properly converted to pre-amp voltage.

    Assumes 10V max, 100 MHz max settings on the V2F100

    """
    chamber = sim_ion_chamber
    # Set the necessary dependent signals
    chamber.counts_per_volt_second = 10e6  # 100 Mhz / 10 V
    chamber.counts.sim_put(int(1.3e7))  # 1.3 V
    chamber.frequency.sim_put(int(10e6))  # 10 MHz clock
    chamber.clock_ticks.sim_put(1e7)  # 1 second @ 10 MHz
    # Check the volts answer
    assert chamber.volts.get() == 1.30


def test_amps_signal(sim_ion_chamber):
    """Test that scaler tick counts get properly converted to ion chamber current."""
    chamber = sim_ion_chamber
    # Set the necessary dependent signals
    chamber.counts_per_volt_second = 10e6  # 100 Mhz / 10 V
    chamber.counts.sim_put(int(13e6))  # 1.3V
    chamber.frequency.sim_put(int(10e6))  # 10 MHz clock
    chamber.clock_ticks.sim_put(1e7)  # 10 MHz clock
    chamber.preamp.sensitivity_value.put(4)  # "20"
    chamber.preamp.sensitivity_unit.put(2)  # "µA/V"
    # Make sure it ignores the offset if it's off
    chamber.preamp.offset_on.put("OFF")
    chamber.preamp.offset_value.put("2")  # 2
    chamber.preamp.offset_unit.put("uA")  # µA
    # Check the current answer
    assert chamber.amps.get() == pytest.approx(2.6e-5)


def test_amps_signal_with_offset(sim_ion_chamber):
    """Test that the scaler tick counts get properly converted to pre-amp voltage."""
    chamber = sim_ion_chamber
    # Set the necessary dependent signals
    chamber.counts.sim_put(int(1.3e7))  # 1.3V
    chamber.counts_per_volt_second = 10e6  # 100 Mhz / 10 V
    chamber.clock_ticks.sim_put(1e7)  # 10 MHz clock
    chamber.frequency.sim_put(int(10e6))  # 10 MHz clock
    chamber.preamp.sensitivity_value.put(4)  # "20"
    chamber.preamp.sensitivity_unit.put(2)  # "µA/V"
    chamber.preamp.offset_on.put("ON")
    chamber.preamp.offset_sign.put("-")
    chamber.preamp.offset_value.put("2")  # 2
    chamber.preamp.offset_unit.put("uA")  # µA
    # Check the current answer
    assert chamber.amps.get() == pytest.approx(2.8e-5)


def test_voltmeter_amps_signal(sim_ion_chamber):
    """Test that the voltmeter voltage gets properly converted to ion
    chamber current.

    """
    chamber = sim_ion_chamber
    # Set the necessary dependent signals
    chamber.voltmeter.volts.sim_put(1.3)  # 1.3V
    chamber.preamp.sensitivity_value.put(4)  # "20"
    chamber.preamp.sensitivity_unit.put(2)  # "µA/V"
    # Make sure it ignores the offset if it's off
    chamber.preamp.offset_on.put("OFF")
    chamber.preamp.offset_value.put("2")  # 2
    chamber.preamp.offset_unit.put("uA")  # µA
    # Check the current answer
    assert chamber.voltmeter.amps.get() == pytest.approx(2.6e-5)


# def test_voltmeter_name(sim_ion_chamber):
#     chamber = sim_ion_chamber
#     assert chamber.voltmeter.description.get() != "Icake"
#     # Change the ion chamber name, and see if the voltmeter name updates
#     chamber.description.put("Icake")
#     assert chamber.voltmeter.description.get() == "Icake"


def test_offset_pv(sim_registry):
    """Check that the device handles the weird offset numbering scheme.

    Net count PVs in the scaler go as

    - 25idcVME:3820:scaler1_netA.B
    - 25idcVME:3820:scaler1_netA.C
    - etc.

    but the offset PVs go
    - 25idcVME:3820:scaler1_offset0.B
    - ...
    - 25idcVME:3820:scaler1_offset0.D
    - 25idcVME:3820:scaler1_offset1.A
    - ...

    """
    channel_suffixes = [
        (2, "offset0.B"),
        (3, "offset0.C"),
        (4, "offset0.D"),
        (5, "offset1.A"),
        (6, "offset1.B"),
        (7, "offset1.C"),
        (8, "offset1.D"),
        (9, "offset2.A"),
        (10, "offset2.B"),
        (11, "offset2.C"),
        (12, "offset2.D"),
    ]
    for ch_num, suffix in channel_suffixes:
        ic = ion_chamber.IonChamber(
            prefix="scaler_ioc:", ch_num=ch_num, name=f"ion_chamber_{ch_num}"
        )
        assert ic.offset.pvname == f"scaler_ioc:scaler1_{suffix}", f"channel {ch_num}"


def test_flyscan_kickoff(sim_ion_chamber):
    flyer = sim_ion_chamber
    flyer.num_bins.set(10)
    status = flyer.kickoff()
    flyer.acquiring.set(1)
    status.wait()
    assert status.success
    assert status.done
    # Check that the device was properly configured for fly-scanning
    assert flyer.erase_start._readback == 1
    assert flyer.timestamps == []
    # Check that timestamps get recorded when new data are available
    flyer.current_channel.set(1).wait()
    assert flyer.timestamps[0] == pytest.approx(time.time())


def test_flyscan_complete(sim_ion_chamber):
    flyer = sim_ion_chamber
    # Run the complete method
    status = flyer.complete()
    status.wait()
    # Check that the detector is stopped
    assert flyer.stop_all._readback == 1


def test_flyscan_collect(sim_ion_chamber):
    flyer = sim_ion_chamber
    name = flyer.net_counts.name
    flyer.start_timestamp = 988.0
    # Make fake fly-scan data
    sim_data = np.zeros(shape=(8000,))
    sim_data[:6] = [3, 5, 8, 13, 2, 33]
    flyer.mca.spectrum._readback = sim_data
    sim_times = np.asarray([12.0e7, 4.0e7, 4.0e7, 4.0e7, 4.0e7, 4.0e7])
    flyer.mca_times.spectrum._readback = sim_times
    flyer.frequency.set(1e7).wait()
    # Ignore the first collected data point because it's during taxiing
    expected_data = sim_data[1:]
    # The real timestamps should be midway between PSO pulses
    flyer.timestamps = [1000, 1004, 1008, 1012, 1016, 1020]
    expected_timestamps = [1002.0, 1006.0, 1010.0, 1014.0, 1018.0]
    payload = list(flyer.collect())
    # Confirm data have the right structure
    for datum, value, timestamp in zip(payload, expected_data, expected_timestamps):
        assert datum == {
            "data": {name: [value]},
            "timestamps": {name: [timestamp]},
            "time": timestamp,
        }


def test_default_time_signal(sim_ion_chamber):
    assert sim_ion_chamber.default_time_signal is sim_ion_chamber.exposure_time


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import pytest
from ophyd.sim import instantiate_fake_device

from haven.instrument.motor import HavenMotor, load_motors
from haven.instrument.motor_flyer import MotorFlyer


@pytest.fixture()
def mocked_device_names(mocker):
    # Mock the caget calls used to get the motor name
    async def resolve_device_names(defns):
        for defn, name in zip(defns, ["SLT V Upper", "SLT V Lower", "SLT H Inbound"]):
            defn["name"] = name

    mocker.patch(
        "haven.instrument.motor.resolve_device_names", new=resolve_device_names
    )


@pytest.fixture()
def motor(sim_registry):
    m1 = instantiate_fake_device(HavenMotor, name="m1")
    m1.user_setpoint._use_limits = False
    return m1


@pytest.mark.asyncio
async def test_load_vme_motors(sim_registry, mocked_device_names):
    # Load the Ophyd motor definitions
    await load_motors()
    # Were the motors imported correctly
    motors = list(sim_registry.findall(label="motors"))
    assert len(motors) == 3
    # assert type(motors[0]) is HavenMotor
    motor_names = [m.name for m in motors]
    assert "SLT_V_Upper" in motor_names
    assert "SLT_V_Lower" in motor_names
    assert "SLT_H_Inbound" in motor_names
    # Check that the IOC name is set in labels
    motor1 = sim_registry.find(name="SLT_V_Upper")
    assert "VME_crate" in motor1._ophyd_labels_


@pytest.mark.asyncio
async def test_skip_existing_motors(sim_registry, mocked_device_names):
    """If a motor already exists from another device, don't add it to the
    motors group.

    """
    # Create an existing fake motor
    m1 = HavenMotor("255idVME:m1", name="kb_mirrors_horiz_upstream", labels={"motors"})
    # Load the Ophyd motor definitions
    await load_motors()
    # Were the motors imported correctly
    motors = list(sim_registry.findall(label="motors"))
    print([m.prefix for m in motors])
    assert len(motors) == 3
    motor_names = [m.name for m in motors]
    assert "kb_mirrors_horiz_upstream" in motor_names
    assert "SLT_V_Upper" in motor_names
    assert "SLT_V_Lower" in motor_names
    # Check that the IOC name is set in labels
    motor1 = sim_registry.find(name="SLT_V_Upper")
    assert "VME_crate" in motor1._ophyd_labels_


def test_motor_signals():
    m = HavenMotor("motor_ioc", name="test_motor")
    assert m.description.pvname == "motor_ioc.DESC"
    assert m.tweak_value.pvname == "motor_ioc.TWV"
    assert m.tweak_forward.pvname == "motor_ioc.TWF"
    assert m.tweak_reverse.pvname == "motor_ioc.TWR"
    assert m.soft_limit_violation.pvname == "motor_ioc.LVIO"


def test_motor_flyer(motor):
    """Check that the haven motor implements the flyer interface."""
    assert motor is not None
    assert isinstance(motor, MotorFlyer)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import logging
from typing import Sequence

from ophyd import ADComponent as ADCpt
from ophyd import CamBase, EpicsSignal, Kind
from ophyd.areadetector.plugins import (
    ImagePlugin_V34,
    OverlayPlugin_V34,
    PvaPlugin_V34,
    ROIPlugin_V34,
)

from .. import exceptions
from .._iconfig import load_config
from .area_detector import (  # noqa: F401
    AsyncCamMixin,
    DetectorBase,
    HDF5FilePlugin,
    SimDetector,
    SingleImageModeTrigger,
    StatsPlugin_V34,
    TIFFFilePlugin,
)
from .device import make_device

log = logging.getLogger(__name__)


__all__ = ["AravisDetector", "load_cameras"]


class AravisCam(AsyncCamMixin, CamBase):
    gain_auto = ADCpt(EpicsSignal, "GainAuto")
    acquire_time_auto = ADCpt(EpicsSignal, "ExposureAuto")


class AravisDetector(SingleImageModeTrigger, DetectorBase):
    """
    A gige-vision camera described by EPICS.
    """

    _default_configuration_attrs = (
        "cam",
        "hdf",
        "stats1",
        "stats2",
        "stats3",
        "stats4",
    )
    _default_read_attrs = ("cam", "hdf", "stats1", "stats2", "stats3", "stats4")

    cam = ADCpt(AravisCam, "cam1:")
    image = ADCpt(ImagePlugin_V34, "image1:")
    pva = ADCpt(PvaPlugin_V34, "Pva1:")
    overlays = ADCpt(OverlayPlugin_V34, "Over1:")
    roi1 = ADCpt(ROIPlugin_V34, "ROI1:", kind=Kind.config)
    roi2 = ADCpt(ROIPlugin_V34, "ROI2:", kind=Kind.config)
    roi3 = ADCpt(ROIPlugin_V34, "ROI3:", kind=Kind.config)
    roi4 = ADCpt(ROIPlugin_V34, "ROI4:", kind=Kind.config)
    stats1 = ADCpt(StatsPlugin_V34, "Stats1:", kind=Kind.normal)
    stats2 = ADCpt(StatsPlugin_V34, "Stats2:", kind=Kind.normal)
    stats3 = ADCpt(StatsPlugin_V34, "Stats3:", kind=Kind.normal)
    stats4 = ADCpt(StatsPlugin_V34, "Stats4:", kind=Kind.normal)
    stats5 = ADCpt(StatsPlugin_V34, "Stats5:", kind=Kind.normal)
    hdf = ADCpt(HDF5FilePlugin, "HDF1:", kind=Kind.normal)
    # tiff = ADCpt(TIFFFilePlugin, "TIFF1:", kind=Kind.normal)


def load_cameras(config=None) -> Sequence[DetectorBase]:
    """Create co-routines for loading cameras from config files.

    Returns
    =======
    coros
      A set of co-routines that can be awaited to load the cameras.

    """
    if config is None:
        config = load_config()
    # Get configuration details for the cameras
    device_configs = {
        k: v
        for (k, v) in config["camera"].items()
        if hasattr(v, "keys") and "prefix" in v.keys()
    }
    # Load each camera
    devices = []
    for key, cam_config in device_configs.items():
        class_name = cam_config.get("device_class", "AravisDetector")
        camera_name = cam_config.get("name", key)
        description = cam_config.get("description", cam_config.get("name", key))
        DeviceClass = globals().get(class_name)
        # Check that it's a valid device class
        if DeviceClass is None:
            msg = f"camera.{key}.device_class={cam_config['device_class']}"
            raise exceptions.UnknownDeviceConfiguration(msg)
        # Create the device object
        devices.append(
            make_device(
                DeviceClass=DeviceClass,
                prefix=f"{cam_config['prefix']}:",
                name=camera_name,
                description=description,
                labels={"cameras", "detectors"},
            )
        )
    return devices


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from unittest import mock

import pytest

from haven.instrument import xia_pfcu
from haven.instrument.xia_pfcu import (
    PFCUFilter,
    PFCUFilterBank,
    PFCUShutter,
    ShutterStates,
    load_xia_pfcu4s,
)


@pytest.fixture()
def shutter(xia_shutter):
    yield xia_shutter


@pytest.fixture()
def shutter_bank(xia_shutter_bank):
    yield xia_shutter_bank


def test_shutter_factory():
    """Check that a shutter device is created if requested."""
    filterbank = PFCUFilterBank(
        prefix="255id:pfcu0:", name="filter_bank_0", shutters=[(2, 3)]
    )  #
    assert filterbank.prefix == "255id:pfcu0:"
    assert filterbank.name == "filter_bank_0"
    assert hasattr(filterbank, "shutters")
    assert isinstance(filterbank.shutters.shutter_0, PFCUShutter)
    assert hasattr(filterbank, "shutters")
    assert isinstance(filterbank.filters.filter1, PFCUFilter)
    assert filterbank.filters.filter1.prefix == "255id:pfcu0:filter1"
    assert isinstance(filterbank.filters.filter4, PFCUFilter)
    assert not hasattr(filterbank.filters, "filter2")
    assert not hasattr(filterbank.filters, "filter3")


def test_pfcu_shutter_signals(shutter):
    # Check initial state
    assert shutter.top_filter.setpoint.get() == 0
    assert shutter.bottom_filter.setpoint.get() == 0


def test_pfcu_shutter_readback(shutter):
    # Set the shutter position
    readback = shutter.parent.parent.readback
    readback._readback = 0b0010
    readback._run_subs(sub_type=readback._default_sub)
    # Check that the readback signal gets updated
    assert shutter.readback.get() == ShutterStates.OPEN


def test_pfcu_shutter_bank_mask(shutter_bank):
    """A bit-mask used for determining how to set the filter bank."""
    shutter = shutter_bank.shutters.shutter_0
    assert shutter.setpoint.top_mask() == 0b0001
    assert shutter.setpoint.bottom_mask() == 0b0010


def test_pfcu_shutter_open(shutter_bank):
    """If the PFCU filter bank is available, open both blades simultaneously."""
    shutter = shutter_bank.shutters.shutter_0
    # Set the other filters on the filter bank
    shutter_bank.readback._readback = 0b0100
    # Open the shutter, and check that the filterbank was set
    shutter.setpoint.set(ShutterStates.OPEN).wait(timeout=1)
    assert shutter_bank.setpoint.get() == 0b0110


def test_pfcu_shutter_close(shutter_bank):
    """If the PFCU filter bank is available, open both blades simultaneously."""
    shutter = shutter_bank.shutters.shutter_0
    # Set the other filters on the filter bank
    shutter_bank.readback._readback = 0b0100
    # Open the shutter, and check that the filterbank was set
    shutter.setpoint.set(ShutterStates.CLOSED).wait(timeout=1)
    assert shutter_bank.setpoint.get() == 0b0101


def test_load_filters(monkeypatch):
    # Simulate the function for making the device
    # works around a bug due to the use of __new__ to make a factory
    device_maker = mock.MagicMock()
    monkeypatch.setattr(xia_pfcu, "make_device", device_maker)
    # Call the code under test
    load_xia_pfcu4s()
    # Check that the fake ``make_device`` function was called properly
    assert device_maker.call_count == 2
    device_maker.assert_called_with(
        PFCUFilterBank,
        labels={"filter_banks"},
        name="filter_bank1",
        prefix="255idc:pfcu1:",
        shutters=[[3, 4]],
    )
    # Make a device with these arguments
    call_args = device_maker.call_args
    device = call_args.args[0](**call_args.kwargs)
    # Check that the filters have the right PVs
    filters = [device.filters.filter1, device.filters.filter2]
    assert isinstance(filters[0], PFCUFilter)
    # Check that the shutter object was created
    shutter = device.shutters.shutter_0
    assert isinstance(shutter, PFCUShutter)
    assert shutter.top_filter.material.pvname == "255idc:pfcu1:filter3_mat"


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2024, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------
